"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 04.05.2022
import time
import random
import numpy as np

from qrisp.core import QuantumSession, QuantumVariable
from qrisp.circuit import RYGate, RXGate, transpile, HGate, ZGate, PGate, XGate, MCXGate
from sympy import Symbol, simplify


def test_abstract_parameters():
    n = 3
    ctrl_qv = QuantumVariable(n)
    target_qv = QuantumVariable(1)

    phi = Symbol("phi")
    multi_cp_gate_0 = PGate(phi).control(n)

    xi = Symbol("xi")
    multi_cp_gate_1 = PGate(-xi).control(n)

    # x(ctrl_qv)
    ctrl_qv.qs.append(multi_cp_gate_0, ctrl_qv.reg + target_qv.reg)
    ctrl_qv.qs.append(multi_cp_gate_1, ctrl_qv.reg + target_qv.reg)

    subs_dic = {xi: 2, phi: 3}
    print(target_qv.get_measurement(subs_dic=subs_dic))

    assert not str(target_qv.get_measurement(subs_dic=subs_dic)) == "{'1': 5.0}"
    assert str(target_qv.get_measurement(subs_dic=subs_dic)) == "{'0': 1.0}"

    ###################
    unitary = simplify(ctrl_qv.qs.get_unitary())

    for i in range(unitary.shape[0]):
        for j in range(unitary.shape[1]):
            if i != j:
                assert unitary[i, j] == 0

    #
    #
    # unitary = Matrix(unitary)
    # init_session()

    ###################
    use_qiskit = False

    if use_qiskit:
        from qiskit.circuit import Parameter, QuantumCircuit
    else:
        from qrisp.circuit import QuantumCircuit
        from sympy import Symbol as Parameter

    n = 1000
    param_name_list = ["p_" + str(i) for i in range(n)]
    parameter_list = [Parameter(x) for x in param_name_list]

    qc = QuantumCircuit(n + 1)

    for i in range(n):
        qc.cx(i, i + 1)
        qc.p(parameter_list[i], i)

    m = 100

    start_time = time.time()
    for i in range(int(m)):
        param_values = [
            random.randint(0, 100) / 100 * 2 * np.pi for j in range(len(parameter_list))
        ]
        qc.bind_parameters(
            {parameter_list[j]: param_values[j] for j in range(len(parameter_list))}
        )

    duration = time.time() - start_time
    # print("Took " + str(duration/m) + " per binding iteration")



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np
from qrisp.misc import array_as_int, gate_wrap, int_as_array
from qrisp.circuit.quantum_circuit import QuantumCircuit

use_gray_code = False
try:
    from ffht import fht as fwht

    fwht(np.array([0, 1, 0, 1]))
    # print("FALCONN fwht import successfull")
except:
    # print("FALCONN fwht import failed")

    def fwht(a):
        """In-place Fast Walsh–Hadamard Transform of array a."""
        h = 1
        while h < len(a):
            for i in range(0, len(a), h * 2):
                for j in range(i, i + h):
                    x = a[j]
                    y = a[j + h]
                    a[j] = x + y
                    a[j + h] = x - y
            h *= 2


def hamming_d(i, j, bit):
    return sum((int_as_array(i, bit) + int_as_array(j, bit)) % 2)


# Solution of the Hamming TSP
# Based on having two salesmen who always visit the place which is closest to them
# and which are reunified at the end of their travel
# The solution is now the path of the first one + the reversed path of the second
def hamming_tsp(location_list, bit_amount):
    if use_gray_code or len(location_list) == 2 ** (bit_amount - 1) + 1:
        return gray_code(max(bit_amount - 1, 1))

    # print(location_list)
    if len(location_list) == 1:
        return list(location_list)

    if len(location_list) == 2:
        temp = list(location_list)
        temp.sort()
        # print([temp[1]] + hamming_movement(temp[1], temp[0], bit_amount))
        return [temp[1]] + hamming_movement(temp[1], temp[0], bit_amount)

    # Copy list in order not to change the input
    location_list = list(location_list)

    # Create log-lists for the salesman pathes
    salesman_1 = [min(location_list)]
    location_list.pop(np.argmin(location_list))

    salesman_2 = [min(location_list)]
    location_list.pop(np.argmin(location_list))

    # Keep track of some stats
    empty_steps = 0

    while len(location_list):
        # Move first salesman

        # Find closest location
        best_distance = np.inf
        for i in range(len(location_list)):
            trial_distance = hamming_d(
                salesman_1[-1], location_list[i], bit_amount
            )  # *len(location_list)
            # trial_distance += hamming_d(salesman_1[-1],
            # salesman_2[-1], n)/len(location_list)

            if trial_distance < best_distance:
                best_distance = trial_distance
                best_candidate_index = i

        # Count steps which do not visit a relevant parity operator
        empty_steps += (
            hamming_d(salesman_1[-1], location_list[best_candidate_index], bit_amount)
            - 1
        )

        # Remove target location from the location list
        target_location = location_list.pop(best_candidate_index)

        # Move salesman
        salesman_1 += hamming_movement(salesman_1[-1], target_location, bit_amount)

        # End procedure if the location list is empty now
        if len(location_list) == 0:
            break

        # Move second salesman (same steps as first salesman)

        best_distance = np.inf
        for i in range(len(location_list)):
            trial_distance = hamming_d(
                salesman_2[-1], location_list[i], bit_amount
            )  # *len(location_list)
            # trial_distance += hamming_d(salesman_1[-1],
            # location_list[i], n)#/len(location_list)

            if trial_distance < best_distance:
                best_distance = trial_distance
                best_candidate_index = i

        empty_steps += (
            hamming_d(salesman_2[-1], location_list[best_candidate_index], bit_amount)
            - 1
        )
        target_location = location_list.pop(best_candidate_index)

        salesman_2 += hamming_movement(salesman_2[-1], target_location, bit_amount)

    # Unify salesmen
    salesman_1 += hamming_movement(salesman_1[-1], salesman_2[-1], bit_amount)

    salesman_2.pop(-1)

    # Construct solution (first salesman + second salesman reversed)
    solution = salesman_1 + salesman_2[::-1]

    # print("Reunion cost:")
    # print(hamming_d(salesman_1[-1], salesman_2[-1], n))
    # empty_steps += hamming_d(salesman_1[-1], salesman_2[-1], n) - 1
    # print("Percentage of empty steps")
    # print(empty_steps/parity_amount*100)
    # print(solution)

    if 2 ** max(bit_amount - 1, 1) < len(solution):
        return gray_code(max(bit_amount - 1, 1))

    return solution


def hamming_movement(init, target, bit_amount):
    salesman = [init]
    distance = hamming_d(init, target, bit_amount)
    while distance:
        distance_array = (
            int_as_array(salesman[-1], bit_amount) + int_as_array(target, bit_amount)
        ) % 2
        movement_index = np.argmax(distance_array)
        movement_array = np.zeros(distance_array.shape)
        movement_array[movement_index] = 1

        salesman.append(
            array_as_int((int_as_array(salesman[-1], bit_amount) + movement_array) % 2)
        )
        distance = hamming_d(salesman[-1], target, bit_amount)
    # print((init, target))
    # print(salesman[1:])
    return salesman[1:]


# Solves the hamming tsp problem i.e. finds a path which visits each
# corner of the n-D unit cube exactly once and ends up in the origin again
def gray_code(n):
    # The key observation is that we can solve the problem for the n-D cube, if we solve
    # it for the (n-1)-D cube such that the path starts at bin(0) and ends in
    # bin(2**(n-2)). We then concantenate this solution in the "lower half" of the n-D
    # cube with the reversed solution in the upper half applied

    # This concantenation will again give a solution which starts at 0 and ends in
    # bin(2**((n+1)-2)). We thus have a recursion

    # Example for the 3D cube:
    # A possible solution for the 2D cube is [0,1,3,2]

    # The reversal is [2,3,1,0] which is [6,7,5,4] on the upper half
    # Concantenating gives [0,1,3,2,6,7,5,4] ie. a solution which starts in 0
    # and ends in 2**(3-1) = 4

    if n < 1:
        return []
    result_list = []

    if n == 1:
        return [1, 0]

    temp_list = gray_code(n - 1)

    result_list += [2 ** (n - 1) + k for k in temp_list[::-1]]
    result_list += temp_list

    return result_list


# This function calculates the phases to put after each parity operator.
# A given state will recieve then recieve the phase phi_i if it is true under the parity
# operator i. As each state is true on a unique combination of parity operators each
# state will recieve a unique phase. The total phase of any state is the sum of all the
# parity operators it is True on. If we now want to get a certain constellation of total
# phases. This gives us a system of linear equations, where the matrix is given by the
# parity matrix


# Target_phases is a list of phases , where the i-th entry describes the phase the
# i-state should pick up. Note that the phases should be given as numbers between 0 and
# 2 where 2 describes a phase of 2pi. This helps to reduce the amount of float errors.
def gen_phase_shifts(target_phases):
    try:
        temp = np.array(target_phases, dtype=np.longdouble) / (len(target_phases) / 2)
    except TypeError:
        temp = [x / (len(target_phases) / 2) for x in target_phases]

    # Transform
    fwht(temp)

    spectrum = temp

    return spectrum


# Solves the hamming tsp and turns it into a sequence of CNOT gates,
# all acting on the same qubit
# Returns a list of tuples indicating CNOT gates and the sequence of
# parity operators traversed
def single_qb_traversal(locations, bit_amount):
    # Solve Hamming TSP to get traversal sequence
    tr_seq = hamming_tsp(locations, bit_amount)

    # Make sure 0 appears last
    if tr_seq[-1] != 0:
        tr_seq = tr_seq + [0]

    if tr_seq[0] != 0:
        tr_seq = [0] + tr_seq

    # Turn into binaries
    tr_seq_bin = [int_as_array(x, bit_amount) for x in tr_seq]

    # Sequence of control bits.
    # The first control bit is the operation bit
    # This indicates later in the synthesis function, that
    # the unmodified parity operator has been loaded into the operation bit
    control_seq = [bit_amount - 1]

    # Now successively calculate Hamming distance vector
    # and determine the non-zero coordinate
    temp = tr_seq[0]
    for i in range(1, len(tr_seq)):
        temp = (tr_seq_bin[i - 1] + tr_seq_bin[i]) % 2
        temp = temp[::-1]

        control_seq.append(np.argmax(temp))

    return control_seq, tr_seq


# Similar to single_qb_traversal but this function also allows traversal
# of Parity operators in more than one qubit
# Returns a list of tuples indicating the applied CNOT gates
# and a list of integers indicating the sequence of traversed parity operators
def multi_qb_traversal(locations, bit_amount):
    # Copy locations list (in oder to prevent modification)

    locations = list(locations)

    # Sort list
    locations.sort()
    locations = locations[::-1]

    # Sort into a dictionary, where the keys are the operation qubits
    # and the values are list of parity operators that have to be traversed
    # on this qubit.
    single_qb_traversal_locations = {len(bin(x)) - 3: [] for x in locations}

    for x in locations:
        single_qb_traversal_locations[len(bin(x)) - 3].append(x)

    # Set up lists to be filled
    cnot_seq = []
    p_op_seq = []

    # Iterate through the qubits which are operated on
    for operation_bit in single_qb_traversal_locations.keys():
        # Convert the traversal locations into single qubit traversal routes
        # (for instance 4,5,7, is the traversal 0,1,3 on the qubit with significance 4)
        if operation_bit != 0:
            single_qb_locations = [
                x % 2 ** (operation_bit)
                for x in single_qb_traversal_locations[operation_bit]
            ]
        else:
            single_qb_locations = single_qb_traversal_locations[operation_bit]

        # Find single qubit traversal routes
        control_seq, p_op_temp = single_qb_traversal(
            [0] + single_qb_locations, operation_bit + 1
        )

        # Append to CNOT list
        cnot_seq += [(x, operation_bit) for x in control_seq]

        # Append to traversed parity operator list
        if operation_bit != 0:
            p_op_seq += [x + 2 ** (operation_bit) for x in p_op_temp]
        else:
            p_op_seq += p_op_temp

    return cnot_seq, p_op_seq


compiled_gates = {}
compiled_pt_gates = {}


def gray_synth_gate(target_phases, phase_tolerant=False):
    target_phases_id = str(list(target_phases))

    if not phase_tolerant:
        if target_phases_id in compiled_gates:
            temp = compiled_gates[target_phases_id]
            return temp
    else:
        if target_phases_id in compiled_pt_gates:
            temp = compiled_pt_gates[target_phases_id]
            return temp

    bit_amount = int(np.log2(len(target_phases)))

    qc = QuantumCircuit(bit_amount)

    # Generate phase shifts
    phase_shifts = gen_phase_shifts(target_phases)

    # Set global phase to 0
    phase_shifts[0] = 0

    # Find out about non-zero parity operators that need to be traversed
    locations = []
    for i in range(len(phase_shifts)):
        try:
            if abs(phase_shifts[i]) < 1e-16:
                continue
        except TypeError:
            pass

        locations.append(i)

    # Generate CNOT list
    cnot_list, p_op_seq = multi_qb_traversal(locations, bit_amount)
    qb_list = qc.qubits

    # Apply CNOT list and add phases
    for i in range(len(cnot_list)):
        if phase_tolerant is True and cnot_list[i][1] != bit_amount - 1:
            # If we are in the phase tolerant mode we don't need to apply a phase here,
            # because we are only interested in the phase DIFFERENCE of the 0 and 1
            # state of the output qubit. As these phase shifts dont act on the output
            # qubit they do not change the phase difference and can therefore be ignored
            break

        # Set some aliases to keep the code readable
        operation_qubit = qb_list[cnot_list[i][1]]
        control_qubit = qb_list[cnot_list[i][0]]

        # Apply CNOT gate
        if cnot_list[i][0] != cnot_list[i][1]:
            control_qubit = qb_list[cnot_list[i][0]]
            qc.cx(control_qubit, operation_qubit)

        # Apply phase shifts
        if phase_shifts[p_op_seq[i]] == 0:
            continue
        qc.p(-phase_shifts[p_op_seq[i]], operation_qubit)
        phase_shifts[p_op_seq[i]] = 0

    from qrisp.core import reduce_depth

    if len(qc.qubits) < 8:
        try:
            qc = reduce_depth(qc)
        except TypeError:
            pass

    qc.qubits = qc.qubits[::-1]

    res = qc.to_gate("gray_phase_gate")
    res.target_phases = target_phases
    res.phase_tolerant = phase_tolerant

    if not phase_tolerant:
        compiled_gates[target_phases_id] = res
    else:
        compiled_pt_gates[target_phases_id] = res

    res.permeability = {i: True for i in range(len(qc.qubits))}
    res.is_qfree = True

    return res


# Function apply gray synthesis to quantum variable qv


# Target_phases is a list of phases , where the i-th entry describes the phase the
# i-state should pick up. Note that the phases should be given as numbers between 0 and
# 2 where 2 describes a phase of 2pi. This helps to reduce the amount of float errors.
@gate_wrap(permeability="full", is_qfree=True)
def gray_phase_synth(qv, target_phases, phase_tolerant=False):
    qv.qs.append(gray_synth_gate(target_phases, phase_tolerant), qv.reg)


# Similar function as the above but takes a QuantumCircuit and a qubit list as arguments
# instead
def gray_phase_synth_qb_list(qc, qb_list, target_phases, phase_tolerant=False):
    qc.append(gray_synth_gate(target_phases, phase_tolerant), qb_list)


# Function to use gray synthesis for logic synthesis
# The general idea is to apply gray synthesis to the input variable and the output
# variables. The output variable however is enclosed H gates which means that it will be
# |1> if we synthesized the |-> state and |0> if we synthesized the |+> state.
# We choose the phases for the states which have a 0 in the output variable to be 0
# and the phases for the states which have a 1 in the output variable according to
# wether the truth table to be synthesized requires a 1 or a 0.

# An additional perk of this function is, that it support phase tolerant synthesis
# This means that we are tolerant regarding the phase the out will have, i.e. if the
# truth table for a given state says 1 and we activate the phase_tolerant option
# we will get the state exp(i*phi)*|1> so a one but with a phase which might vary
# for different input states. This tolerance regarding the phase allows us to skip half
# of the synthesis procedure making it very cheap regarding the gate count


# input_var is the quantum variable which contains the input
# output_var is the quantum variable where the output should be synthesized
# Note that output_var doesn't need to be of size one - the next parameter
# qb_nr indicates on which qubit of output_var the synthesis should take place
# Finally tt is the truth table to be synthesized
@gate_wrap(is_qfree=True, permeability=[0])
def gray_logic_synth_single_qb(input_var, output_var, qb_nr, tt, phase_tolerant=False):
    if input_var.size != tt.bit_amount:
        raise Exception(
            "Input variable does not contain enough qubits to encode truth table"
        )

    if tt.shape[1] != 1:
        raise Exception("This function can only encode single column truth tables")

    # Generate target phases such that all the states with a 0 in the output qubit
    # get phase 0 and the states with a 1 in the input qubit get a phases of pi
    # if the truth table says 1 or a phase of 0 otherwise
    target_phases = np.array(tt.shape[0] * [0] + list(tt.n_rep[:, 0])) * np.pi

    # Apply h gate to the output qubit in order to get logic states from phases
    from qrisp import h

    h(output_var[qb_nr])
    input_var.qs.append(
        gray_synth_gate(target_phases, phase_tolerant),
        (input_var.reg + [output_var[qb_nr]])[::-1],
    )
    h(output_var[qb_nr])
    return


# This function uses it's single qubit version iteratively to synthesize
# truth tables with more than one column
@gate_wrap(is_qfree=True, permeability=[0])
def gray_logic_synth(input_var, output_var, tt, phase_tolerant=False, lin_solve=False):
    if len(input_var) != tt.bit_amount:
        raise Exception(
            "Input variable does not contain enough qubits to encode truth table"
        )

    if len(output_var) != tt.shape[1]:
        raise Exception(
            "output variable does not contain enough qubits to encode truth table"
        )

    input_var_dupl = input_var.duplicate()
    output_var_dupl = output_var.duplicate(qs=input_var_dupl.qs)

    # Iterate throgh all the truth table columns and synthesize each one separately

    residual_phases = np.zeros(tt.shape[0])

    for i in range(tt.shape[1]):
        gray_logic_synth_single_qb(
            input_var_dupl, output_var_dupl, i, tt.sub_table(i), phase_tolerant=True
        )

        temp = (tt.sub_table(i).n_rep - 1 / 2) * np.pi / 2
        temp = temp.transpose()[0]

        residual_phases += temp

    from qrisp.core import reduce_depth

    qc = input_var_dupl.qs.transpile()

    if len(qc.qubits) < 10 and input_var.size != 1:
        pass
        # qc = reduce_depth(qc)

    input_var_dupl.qs.data = []

    if not phase_tolerant:
        input_var_dupl.reg = input_var_dupl.reg[::-1]

        gray_phase_synth(input_var_dupl, residual_phases)

        input_var_dupl.reg = input_var_dupl.reg[::-1]

    qc.data.extend(input_var_dupl.qs.data)

    input_var.qs.append(qc.to_gate(), input_var.reg + output_var.reg)


# This function uses it's single qubit version iteratively to synthesize
# truth tables with more than one column
@gate_wrap
def gray_logic_synth_qb_list(
    input_qb_list,
    output_qb,
    qs,
    tt,
    phase_tolerant=False,
    inverse=False,
    gate_name=None,
):
    from qrisp import QuantumSession, QuantumVariable

    qs_temp = QuantumSession()

    input_var = QuantumVariable(len(input_qb_list), qs_temp)
    output_var = QuantumVariable(1, qs_temp)

    gray_logic_synth(input_var, output_var, tt, phase_tolerant)

    temp_gate = qs_temp.to_gate()

    if gate_name is not None:
        temp_gate.name = gate_name

    if inverse:
        qs.append(temp_gate.inverse(), input_qb_list + [output_qb])
    else:
        qs.append(temp_gate, input_qb_list + [output_qb])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import h, QuantumFloat, multi_measurement, auto_uncompute, QuantumBool, mcx
from qrisp.quantum_backtracking import QuantumBacktrackingTree
import numpy as np
from sympy import nsimplify, Float

def test_backtracking():
    
    @auto_uncompute    
    def P(tree):
        temp_0 = (tree.h == 0)
        predicate = QuantumBool()
        mcx(list(tree.branch_qa) + list(temp_0), predicate)
        
        return predicate

    @auto_uncompute    
    def Q(tree):
        return (tree.h != tree.max_depth) & (tree.h != (tree.max_depth-1)) & (tree.branch_qa[0] == tree.branch_qa[1])

    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    tree.init_node([])
    res = tree.estimate_phase(4)
    
    mes_res = res.get_measurement()
    
    assert mes_res[0] < 0.25
    
    
    @auto_uncompute    
    def P(tree):
        predicate = QuantumBool()
        mcx(list(tree.branch_qa), predicate)
        return predicate

    @auto_uncompute    
    def Q(tree):
        return QuantumBool()
    
    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    tree.init_node([])
    res = tree.estimate_phase(4)
    
    mes_res = res.get_measurement()
    
    assert mes_res[0] > 0.375
    
    for i in range(1,5):
    
        tree = QuantumBacktrackingTree(i, QuantumFloat(1, name = "branch_qf*"), P, Q)
        
        tree.init_phi(i*[1])
        
        temp_sv_0 = tree.qs.statevector().expand().evalf()
        tree.quantum_step()
        temp_sv_1 = tree.qs.statevector().expand().evalf()
        
        diff = (temp_sv_0 - temp_sv_1).expand().evalf()
        diff_atoms = list(diff.atoms(Float))
        s = sum(diff_atoms)
        assert abs(s) < 1E-4
        
        tree = QuantumBacktrackingTree(i, QuantumFloat(1, name = "branch_qf*"), P, Q)
        tree.init_node(i*[1])
        
        temp_sv_0 = tree.qs.statevector().expand().evalf()
        tree.qstep_diffuser(even = True)
        temp_sv_1 = tree.qs.statevector().expand().evalf()
        
        diff = (temp_sv_0 - temp_sv_1).expand().evalf()
        diff_atoms = list(diff.atoms(Float))
        s = sum(diff_atoms)
        assert abs(s) < 1E-4
    
        
    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    solution = tree.find_solution(precision = 4)
    assert solution == [1,1,1]


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import copy

import numpy as np
from sympy.core.expr import Expr


def adaptive_substitution(expr, subs_dic, precision=10):
    if isinstance(expr, Expr):
        return float(expr.evalf(precision, subs_dic))
    elif isinstance(expr, (float, int)):
        # TO-DO
        return float(expr)
    else:
        # TO-DO
        return expr


# Class that describes an operation which can be performed on a quantum computer
# Example would be an X gate or a measurement
class Operation:
    """
    This class describes operations like quantum gates, measurements or classical logic
    gates. Operation objects do not carry information about which Qubit/Clbits they are
    applied to. This can be found in the Instruction class, which is a combination of an
    Operation object together with its operands.

    Operation objects consist of five basic attributes:

    * ``.name`` : A string identifying the operation
    * ``.num_qubits`` : An integer specifying the amount qubits on which to operate
    * ``.num_clbits`` : An integer specifying the amount of classical bits on which to
      operate.
    * ``.params`` : A list of floats specifying the parameters of the Operation
    * ``.definition`` : A :ref:`QuantumCircuit`. For synthesized (i.e. non-elementary)
      operations, this QuantumCircuit specifies the operation.

    Operation objects together with their Operands can be appended to QuantumCircuits by
    using the :meth:`append <qrisp.QuantumCircuit.append>` method.

    QuantumCircuits can be turned into Operations by using the
    :meth:`to_gate <qrisp.QuantumCircuit.to_gate>` method.

    Examples
    --------

    We create a QuantumCircuit and append a couple of operations

    >>> from qrisp import QuantumCircuit, XGate, CXGate, PGate
    >>> qc = QuantumCircuit(2)
    >>> qc.append(XGate(), 0)
    >>> qc.append(CXGate(), [0,1])
    >>> qc.append(PGate(0.5), 1)
    >>> synthed_op = qc.to_op()
    >>> qc.append(synthed_op, qc.qubits)

    """

    # If only given the operation init_op (which can be portable) a copied instance of
    # this operation will be returned
    def __init__(
        self,
        name=None,
        num_qubits=0,
        num_clbits=0,
        definition=None,
        params=[],
        unitary=None,
        init_op=None,
    ):
        if init_op is not None:
            name = init_op.name
            num_qubits = init_op.num_qubits
            num_clbits = init_op.num_clbits
            params = init_op.params
            definition = init_op.definition

            try:
                unitary = init_op.unitary
            except AttributeError:
                pass
        else:
            if name is None:
                raise Exception("Tried to create a Operation object without name")

        # Name of the operation - this is how the backend behind the interface will
        # identify the operation
        self.name = name

        # Amount of qubits
        self.num_qubits = num_qubits

        # Amount of classical bits
        self.num_clbits = num_clbits

        # List of parameters (also available behind the interface)
        self.params = []

        # Find abstract parameters (ie. sympy expressions and log them)
        self.abstract_params = set([])

        for par in params:
            if isinstance(par, Expr):
                self.abstract_params = self.abstract_params.union(par.free_symbols)
            elif not isinstance(par, (np.floating, np.int32, float, int)):
                raise Exception(
                    f"Tried to create operation with parameters of type {type(par)}"
                )

            self.params.append(par)

        # If a definition circuit is given, this means we are supposed to create a
        # non-elementary operation
        if definition is not None:
            # Copy circuit in order to prevent modification
            # self.definition = QuantumCircuit(init_qc = definition)
            self.definition = definition

            self.abstract_params = self.abstract_params.union(
                definition.abstract_params
            )

        else:
            self.definition = None

        # If a unitary (numpy array) is given, save the unitary
        if unitary is not None:
            self.unitary = unitary

        # These attributes store some information for the uncomputation algorithm
        # Qfree basically means that the unitary is a permutation matrix
        # (up to local phase shifts). Permeability means that this gate commutes with
        # the z operator on a given qubit
        self.is_qfree = None
        self.permeability = {i: None for i in range(self.num_qubits)}

    def copy(self):
        """
        Returns a copy of the Operation object.

        Returns
        -------
        Operation
            The copied operation.

        """

        res = copy.copy(self)
        if self.definition:
            copied_definition = self.definition.copy()
        else:
            copied_definition = None

        res.definition = copied_definition

        return res

    # Method to get the unitary matrix of the operation

    # The parameter decimals has no influence on what is calculated
    # Rounding is usefull here because the floating point errors
    # sometimes make it hard to read the unitary
    def get_unitary(self, decimals=-1):
        """
        Returns the unitary matrix (if applicable) of the Operation as a numpy array.

        Parameters
        ----------
        decimals : int, optional
            Amount of decimals to return. By default, the full precision is returned.

        Raises
        ------
        Exception
            Could not calculate the unitary.

        Returns
        -------
        numpy.ndarray
            The unitary matrix of the Operation.


        Examples
        --------

        >>> from qrisp import CPGate
        >>> import numpy as np
        >>> CPGate(np.pi/2).get_unitary(decimals = 3)
        array([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+1.j]], dtype=complex64)
        """

        if self.name == "barrier":
            return np.eye(2**self.num_qubits)

        # Check if the unitary is already available
        if hasattr(self, "unitary"):
            if decimals == -1:
                return self.unitary
            else:
                return np.round(self.unitary, decimals)

        # If we are dealing with a non-elementary gate, calculate the unitary from
        # the definition circuit
        else:
            if not isinstance(self.definition, type(None)):
                self.unitary = self.definition.get_unitary()
                return self.get_unitary()

            # If no definition circuit is known, raise an error.
            # Note that the get_unitary methods of more specific gate families specified
            # by the inheritors of this class
            else:
                raise Exception("Unitary of operation " + self.name + " not defined.")

    # Method to return the inverse of the given operation. Again, the methods of more
    # specific gate families are specified by the inheritors of this class
    def inverse(self):
        """
        Returns the inverse of this Operation (if applicable).

        Raises
        ------
        Exception
            Tried to invert non-unitary operation.

        Returns
        -------
        Operation
            The daggered operation.

        Examples
        --------

        We invert a phase gate and inspect it's parameters

        >>> from qrisp import PGate
        >>> phase_gate = PGate(0.8)
        >>> phase_gate.inverse().params
        [-0.8]

        """

        # Check if the instruction contains classical bits => operation is not
        # invertible
        if self.num_clbits:
            raise Exception("Tried to invert non-unitary operation")

        # Check if a definition is available and invert it, if so
        if self.definition is not None:
            inverse_circ = self.definition.inverse()
            res = inverse_circ.to_op(name=self.name + "_dg")

        elif self.name == "qb_alloc":
            from qrisp.circuit import QubitDealloc

            res = QubitDealloc()
        elif self.name == "qb_dealloc":
            from qrisp.circuit import QubitAlloc

            res = QubitAlloc()
        # Otherwise raise an error
        else:
            raise Exception("Don't know how to invert Operation " + self.name)

        res.is_qfree = self.is_qfree
        res.permeability = dict(self.permeability)

        return res

    # Method to create a controlled gate
    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):
        """
        Returns the controlled version of this Operation (if applicable).

        Parameters
        ----------
        num_ctrl_qubits : int, optional
            The amount of control qubits. The default is 1.
        ctrl_state : int or str, optional
            The state on which to activate the basis gate. The default is "1111...".
        method : str, optional
            The method for synthesizing the required multi-controlled X gates.
            Available are ``gray`` and ``gray_pt`` and ``auto``.
            Note that "gray_pt" introduces an extra phase (which needs to be uncomputed)
            but is more resource efficient. ``auto`` will be transformed into a more
            efficient ancilla supported version at compile time if used in a
            QuantumSession. The default is ``gray``.

        Raises
        ------
        AttributeError
            Tried to control non-unitary operation.

        Returns
        -------
        Operation
            The controlled operation.

        Examples
        --------

        We control a parametrized X Rotation.

        >>> from qrisp import QuantumCircuit, RXGate
        >>> mcrx_gate = RXGate(0.5).control(3)
        >>> qc = QuantumCircuit(4)
        >>> qc.append(mcrx_gate, qc.qubits)
        >>> print(qc)
        qb_4: ─────■─────
                   │
        qb_5: ─────■─────
                   │
        qb_6: ─────■─────
              ┌────┴────┐
        qb_7: ┤ Rx(0.5) ├
              └─────────┘


        """

        if method is None:
            method = "auto"

        if self.num_clbits != 0:
            raise AttributeError("Tried to control non-unitary operation")
            
        res_num_ctrl_qubits = num_ctrl_qubits
        
        if isinstance(self, PTControlledOperation):
            res_num_ctrl_qubits += len(self.controls)
        
        # Check if the method is phase tolerant
        if method.find("pt") != -1 and res_num_ctrl_qubits != 1:
            
            return PTControlledOperation(
                self, num_ctrl_qubits, ctrl_state=ctrl_state, method=method
            )
        else:
            return ControlledOperation(
                self, num_ctrl_qubits, ctrl_state=ctrl_state, method=method
            )

    # TO-DO implement more robust hashing method
    def __hash__(self):
        return hash(hash(self.name) + hash(tuple(self.params)))

    def is_permeable(self, indices):
        from qrisp.uncomputation import is_permeable

        return is_permeable(self, indices)

    def bind_parameters(self, subs_dic):
        """
        Binds abstract parameters to specified values.

        Parameters
        ----------
        subs_dic : dict
            A dictionary containing the parameters as keys.

        Returns
        -------
        res : Operation
            The Operation with bound parameters.

        Examples
        --------

        We create a phase gate with an abstract parameter and bind it to a specified
        value.

        >>> from qrisp import PGate
        >>> from sympy import Symbol
        >>> phi = Symbol("phi")
        >>> abstract_p_gate = PGate(phi)
        >>> abstract_p_gate.params
        [phi]
        >>> bound_p_gate = abstract_p_gate.bind_parameters({phi : 1.5})
        >>> bound_p_gate.params
        [1.5]
        """

        new_params = []
        for i in range(len(self.params)):
            new_params.append(adaptive_substitution(self.params[i], subs_dic))

        temp = self.params
        self.params = new_params
        res = Operation(init_op=self)
        self.params = temp

        if not isinstance(res.definition, type(None)):
            res.definition = res.definition.bind_parameters(subs_dic)

        return res


# Class to describe 1-Qubit gates as unitaries of the form U(theta, phi, lam) =
# = exp(-1j*phi/2*sigma_z) exp(-1j*theta/2*sigma_y) exp(-1j*lam/2*sigma_z)
# See https://qiskit.org/documentation/stubs/qiskit.circuit.library.U3Gate.html
# for more information
class U3Gate(Operation):
    def __init__(self, theta, phi, lam, name="u3", global_phase=0):
        self.global_phase = global_phase
        # Initialize Operation instance
        super().__init__(
            name=name,
            num_qubits=1,
            num_clbits=0,
            definition=None,
            params=[theta, phi, lam],
        )

        # Set parameters
        self.theta = self.params[0]
        self.phi = self.params[1]
        self.lam = self.params[2]

        if self.name in ["rx", "ry", "rz", "p"]:
            self.params = [sum(self.params)]

            if self.name in ["rz", "p"]:
                self.permeability[0] = True
                self.is_qfree = True
            else:
                self.permeability[0] = False
                self.is_qfree = False

        elif self.name in ["h"]:
            self.params = []

            self.permeability[0] = False
            self.is_qfree = False

    # Specify inversion method
    def inverse(self):
        # The inverse of a product of matrices if the reverted product of the inverses,
        # i.e. (A*B*C)^(-1) = C^-1 * B^-1 * A^-1

        if self.name[-3:] == "_dg":
            new_name = self.name[:-3]
        else:
            new_name = self.name + "_dg"

        # For exponentials of hermitian matrices, the inverse is the hermitean
        # conjugate, which implies that we simply have to negate the parameters
        # theta, phi, lambda
        res = U3Gate(
            -self.theta,
            -self.lam,
            -self.phi,
            name=new_name,
            global_phase=-self.global_phase,
        )

        if self.name == "u3":
            res.name = "u3"

        # These are special gates that require only a single parameter
        if self.name in ["rx", "ry", "rz", "p", "h", "gphase"]:
            res.name = self.name
            res.params = [-par for par in self.params]

        if self.name in ["s", "t", "s_dg", "t_dg"]:
            res.params = []

        if res.is_qfree is not None:
            res.is_qfree = bool(self.is_qfree)

        res.permeability = dict(self.permeability)
        res.is_qfree = bool(self.is_qfree)

        return res

    # Method to calculate the unitary matrix
    def get_unitary(self, decimals=-1):
        if hasattr(self, "unitary"):
            if decimals != -1:
                return np.around(self.unitary, decimals)
            else:
                return self.unitary
        else:
            from qrisp.simulator.unitary_management import u3matrix

            # Generate unitary

            self.unitary = u3matrix(self.theta, self.phi, self.lam, self.global_phase)

            return self.get_unitary(decimals)

    def bind_parameters(self, subs_dic):
        return U3Gate(
            adaptive_substitution(self.theta, subs_dic),
            adaptive_substitution(self.phi, subs_dic),
            adaptive_substitution(self.lam, subs_dic),
            self.name,
            adaptive_substitution(self.global_phase, subs_dic),
        )

    def c_if(self, num_clbits=1, value=-1):
        if value == -1:
            value = 2**num_clbits - 1

        res = self.copy()
        res.num_clbits += 1
        res.name = "c_if_" + res.name
        return res


id_matrix = np.eye(2, dtype=np.complex64)

pauli_x = np.asarray([[0, 1], [1, 0]], dtype=np.complex64)
pauli_y = np.asarray([[0, 0], [0, 0]]) + 1j * np.asarray([[0, -1], [1, 0]])
pauli_z = np.asarray([[1, 0], [0, -1]], dtype=np.complex64)

pi = float(np.pi)


# This class is special for pauli gates. In principle, we could also use the U3Gate
# class, but this could lead to unneccessary floating point errors
class PauliGate(U3Gate):
    def __init__(self, name):
        if name == "x":
            super().__init__(pi, 0, pi)
            self.unitary = pauli_x

            self.is_qfree = True
            self.permeability[0] = False

        elif name == "y":
            super().__init__(pi, pi / 2, pi / 2)
            self.unitary = pauli_y

            self.is_qfree = True
            self.permeability[0] = False

        elif name == "z":
            super().__init__(0, 0, pi)
            self.unitary = pauli_z

            self.is_qfree = True
            self.permeability[0] = True

        else:
            raise Exception("Gate " + name + " is not a Pauli gate")

        self.name = name
        self.params = []

    def inverse(self):
        return PauliGate(self.name)

    def __repr__(self):
        return self.name


# This class describes phase tolerant controlled operations
# Phase tolerant means that
class PTControlledOperation(Operation):
    def __init__(self, base_operation, num_ctrl_qubits=1, ctrl_state=-1, method="auto"):
        # Object which describes the method. Can be a string lr a callable function
        self.method = method

        # QuantumOperation object which describes which action is controlled,
        # i.e. for a CX gate, this would be an X gate
        self.base_operation = base_operation

        # List of control qubits
        self.controls = list(range(num_ctrl_qubits))

        # The control state - can be specified either as a string or as an int
        if ctrl_state == -1:
            self.ctrl_state = num_ctrl_qubits * "1"
        elif isinstance(ctrl_state, int):
            from qrisp.misc import bin_rep

            self.ctrl_state = bin_rep(ctrl_state, num_ctrl_qubits)[::-1]
        else:
            self.ctrl_state = str(ctrl_state)

        # Check if control state specification matches control qubit amount
        if len(self.ctrl_state) != num_ctrl_qubits:
            raise Exception(
                "Specified control state incompatible with given control qubit amount"
            )

        # Now we generate the definition circuit. Note that most of the generation
        # process also applies to the ControlledOperation class, however this class has
        # more specific method (.inverse() for instance),
        # which is why we make this distinction

        # If the base operation is a ControlledOperation, the result
        # PTControlledOperation can be generated by applying the phase tolerant control
        # algorithm to the base gate of the controlled operation in question
        elif isinstance(base_operation, PTControlledOperation):
            if method == "gray":
                method = self.method
            self.__init__(
                base_operation.base_operation,
                num_ctrl_qubits=num_ctrl_qubits + len(base_operation.controls),
                ctrl_state=self.ctrl_state + base_operation.ctrl_state,
                method=method,
            )
            return

        if base_operation.name == "gphase":
            from qrisp import PGate, QuantumCircuit, bin_rep

            definition_circ = QuantumCircuit(num_ctrl_qubits + 1)

            if num_ctrl_qubits > 1:
                temp_gate = PGate(base_operation.params[0]).control(
                    num_ctrl_qubits - 1, ctrl_state=self.ctrl_state[1:], method=method
                )
            else:
                temp_gate = PGate(base_operation.params[0])

            if self.ctrl_state[0] == "0":
                definition_circ.x(-2)

            definition_circ.append(temp_gate, range(num_ctrl_qubits))

            if self.ctrl_state[0] == "0":
                definition_circ.x(-2)

        elif self.base_operation.name == "gray_phase_gate":
            from qrisp.circuit import multi_controlled_gray_circ

            definition_circ, target_phases = multi_controlled_gray_circ(
                self.base_operation, num_ctrl_qubits, self.ctrl_state
            )
            self.target_phases = target_phases
            self.phase_tolerant = False

        elif self.base_operation.name == "swap":
            from qrisp.core import fredkin_qc

            definition_circ = fredkin_qc(num_ctrl_qubits, ctrl_state, method)

        # For the case of a pauli gate with a single control, we insert an extra case
        # since here is no need for any advanced algorithm here and we do not need
        # to apply the phase tolerant naming convention
        elif (
            isinstance(base_operation, PauliGate)
            and num_ctrl_qubits == 1
            and self.ctrl_state == "1"
        ):
            if base_operation.name == "x":
                super().__init__(
                    name="cx", num_qubits=2, num_clbits=0, params=[], unitary=None
                )
                self.permeability = {0: True, 1: False}
            elif base_operation.name == "y":
                super().__init__(
                    name="cy", num_qubits=2, num_clbits=0, params=[], unitary=None
                )
                self.permeability = {0: True, 1: False}
            elif base_operation.name == "z":
                super().__init__(
                    name="cz", num_qubits=2, num_clbits=0, params=[], unitary=None
                )
                self.permeability = {0: True, 1: True}

            self.is_qfree = True
            return

        # In the case of an u3gate as a base operation, we first check, if the method
        # object is callable, we generate the definition circuit from this function.
        # Otherwise we call the algorithm for generating controlled u3 gates
        elif isinstance(base_operation, U3Gate):
            if callable(method):
                definition_circ = method(self, num_ctrl_qubits, self.ctrl_state)
            else:
                from qrisp.circuit.controlled_operations import multi_controlled_u3_circ

                definition_circ = multi_controlled_u3_circ(
                    base_operation,
                    num_ctrl_qubits,
                    ctrl_state=self.ctrl_state,
                    method=method,
                )

        # If the base operation has a definition, we can simply apply the phase tolerant
        # control algorithm to every gate contained in this defintion.
        # This is done in the function multi_controlled_circuit

        elif base_operation.definition:
            from qrisp.circuit import multi_controlled_circuit

            definition_circ = multi_controlled_circuit(
                base_operation.definition,
                num_ctrl_qubits,
                ctrl_state=self.ctrl_state,
                method=method,
            )

        # Raise exception if no possility of synthesizing a controlled game is known
        else:
            raise Exception(
                "Control method for gate " + base_operation.name + " not implemented"
            )

        # Generate gate name
        if num_ctrl_qubits == 1:
            name_prefix = "ptc"
        else:
            name_prefix = "pt" + str(num_ctrl_qubits) + "c"

        Operation.__init__(
            self,
            name=name_prefix + base_operation.name,
            num_qubits=base_operation.num_qubits + num_ctrl_qubits,
            num_clbits=0,
            definition=definition_circ,
            params=base_operation.params,
            unitary=None,
        )

        for i in range(self.num_qubits):
            if i < num_ctrl_qubits:
                self.permeability[i] = True
            else:
                self.permeability[i] = base_operation.permeability[i - num_ctrl_qubits]

        if base_operation.is_qfree is not None:
            self.is_qfree = bool(base_operation.is_qfree)

    def inverse(self):
        # Generate inverse operation by applying the constructor
        # to the inverse base_operation to get the meta-data right

        res = PTControlledOperation(
            self.base_operation.inverse(),
            len(self.controls),
            ctrl_state=self.ctrl_state,
            method=self.method,
        )

        # In order to make sure the definition circuit is also correct,
        # we invert the circuit if its invertible
        try:
            res.definition = self.definition.inverse()
        except AttributeError:
            pass

        if res.is_qfree is not None:
            res.is_qfree = bool(self.is_qfree)

        res.permeability = dict(self.permeability)

        return res

    def bind_parameters(self, subs_dic):
        from copy import copy

        res = copy(self)
        if not isinstance(self.definition, type(None)):
            res.definition = self.definition.bind_parameters(subs_dic)
        res.base_operation = self.base_operation.bind_parameters(subs_dic)

        res.params = []
        for p in self.params:
            if p in subs_dic:
                res.params.append(subs_dic[p])
            else:
                res.params.append(p)

        return res


# Class to describe controlled operation
# Very simlar to phase tolerant operations but with a more specifix naming
# convention, inversion algorithm and unitary generation algorithm
class ControlledOperation(PTControlledOperation):
    def __init__(self, base_operation, num_ctrl_qubits=1, ctrl_state=-1, method="gray"):
        super().__init__(
            base_operation,
            num_ctrl_qubits=num_ctrl_qubits,
            ctrl_state=ctrl_state,
            method=method,
        )

        if num_ctrl_qubits == 1:
            name_prefix = "c"
        else:
            name_prefix = str(len(self.controls)) + "c"

        self.name = name_prefix + base_operation.name

    def inverse(self):
        return ControlledOperation(
            self.base_operation.inverse(),
            len(self.controls),
            ctrl_state=self.ctrl_state,
            method=self.method,
        )

    # For generating the unitary we don't have to generate the unitary by multiplying
    # the gates of the definition circuit.
    # Instead, we can use that the unitary of a controlled operation is
    # the identity matrix apart from the bottom right block matrix, where we find
    # the unitary of the base operation.

    def get_unitary(self, decimals=-1):
        if hasattr(self, "unitary"):
            if decimals != -1:
                return np.around(self.unitary, decimals)
            else:
                return self.unitary
        else:
            from qrisp.simulator.unitary_management import controlled_unitary

            self.unitary = controlled_unitary(self)
            return self.get_unitary(decimals)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import weakref

import numpy as np

from qrisp.circuit import Clbit, QuantumCircuit, Qubit, QubitAlloc, QubitDealloc
from qrisp.core.session_merging_tools import multi_session_merge
from qrisp.misc import get_depth_dic


class QuantumSession(QuantumCircuit):
    """
    The QuantumSession class manages the life cycle of QuantumVariables and enables
    features such as :ref:`QuantumEnvironments <QuantumEnvironment>` or
    :ref:`Uncomputation`. To create a QuantumSession, we call the constructor

    >>> from qrisp import QuantumSession
    >>> qs = QuantumSession()

    To create a QuantumVariable within that QuantumSession, we hand it over to the
    QuantumVariable constructor:

    >>> from qrisp import QuantumVariable
    >>> qv = QuantumVariable(3, qs = qs)

    As an inheritor of the :ref:`QuantumCircuit` class, QuantumSession objects can also
    be used for circuit construction.

    >>> qv.qs.cx(qv[0], qv[1])

    Nevertheless, users are encouraged to use the :ref:`designated gate application
    function <gate_application_functions>` in order to reduce code cluttering.

    >>> from qrisp import cx
    >>> cx(qv[0], qv[1])

    QuantumSessions can be visualized by calling ``print`` on them.

    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
    qv.0: ──■────■──
          ┌─┴─┐┌─┴─┐
    qv.1: ┤ X ├┤ X ├
          └───┘└───┘
    qv.2: ──────────
    <BLANKLINE>
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    If not specified, QuantumVariables will create their own QuantumSession and register
    themselves in it.

    QuantumSessions can be manually merged using the merge function.

    >>> from qrisp import merge
    >>> qs_2 = QuantumSession()
    >>> qs_2 == qs
    False
    >>> merge(qs, qs_2)
    >>> qs == qs_2
    True

    Note that merge also works for QuantumVariables, lists of QuantumSession and lists
    of QuantumVariables.

    If an entangling operation between two QuantumVariables which are registered in
    different QuantumSessions is executed, these QuantumSessions are automatically
    merged. For more details on automatic QuantumSession merging check the
    :ref:`session merging documentation<SessionMerging>`.

    >>> qv_a = QuantumVariable(2)
    >>> qv_b = QuantumVariable(2)
    >>> qv_a.qs == qv_b.qs
    False
    >>> from qrisp import cx
    >>> cx(qv_a[0], qv_b[0])
    >>> qv_a.qs == qv_b.qs
    True

    QuantumSessions can be given a default backend on which to evaluate circuits:

    >>> from qrisp.interface import VirtualQiskitBackend
    >>> qiskit_backend = instantiate_qiskit_backend()
    >>> qs = QuantumSession(backend = VirtualQiskitBackend(qiskit_backend))

    In this piece of code, we assume that the function ``instantiate_qiskit_backend``
    creates a Qiskit backend instance (which could either be the QASM Simulator or a
    real backend). We then hand this to the :ref:`VirtualQiskitBackend` constructor to
    turn it into a Qrisp backend. Now, any measurements of variables that are registered
    in this session will be evaluated on that backend.

    If no backend is given, the backend specified in ``default_backend.py`` will be
    used.

    Note that it is not possible to merge two QuantumSessions with differing,
    non-trivial backends.
    """

    qs_tracker = []

    def __init__(self, backend=None):
        """
        Constructs a QuantumSession

        Parameters
        ----------
        backend : BackendClient, optional
            The backend on which to execute the circuits created by this QuantumSession.
            This choice can be overwritten by specifying a backend in the
            :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>` method of
            QuantumVariable.

        Returns
        -------
        None.

        Examples
        --------

        We create a QuantumSession with the QASM simulator as default backend and
        register a QuantumFloat in it:

        >>> from qiskit import Aer
        >>> qasm_sim = Aer.get_backend("qasm_simulator")
        >>> from qrisp.interface import VirtualQiskitBackend
        >>> vrtl_qasm_sim = VirtualQiskitBackend(qasm_sim)
        >>> from qrisp import QuantumSession, QuantumFloat
        >>> qs = QuantumSession(vrtl_qasm_sim)
        >>> qf = QuantumFloat(4, qs = qs)


        """

        if isinstance(backend, int):
            raise Exception

        super().__init__()

        self.backend = backend

        # Set up list of quantum variables appearing in this session

        self.qv_list = []
        self.deleted_qv_list = []

        # Set up environment stack
        # This list will be filled, once we enter an environment
        self.env_stack = []

        # This list will be filled with variables which are marked for uncomputation
        # Variables will be marked once there is no longer any reference to them apart
        # from the one in qv_list. This is for instance the case with local variables
        # inside a function after the function finished
        self.uncomp_stack = []

        self.qs_tracker.append(weakref.ref(self))

        # This list will contain the QuantumSessions which have been merged into this
        # session. It needs to be tracked in order to also update the shadow sessions
        # when this session is merged into another session.
        self.shadow_sessions = []

    def register_qv(self, qv):
        """
        Method to register QuantumVariables

        Parameters
        ----------
        qv : QuantumVariable
            QuantumVariable to register.

        Raises
        ------
        RuntimeError
            Name of qv is already used in this QuantumSession.

        Returns
        -------
        None.

        """
        if qv.name in [temp_qv.name for temp_qv in self.qv_list + self.deleted_qv_list]:
            raise RuntimeError(
                "Variable name " + str(qv.name) + " already exists in quantum session"
            )

        # Determine amount of required qubits
        req_qubits = qv.size

        # Hand qubits to quantum variable
        qv.reg = self.request_qubits(req_qubits, name=qv.name)

        # Register in the list of active quantum variable
        self.qv_list.append(qv)

    def get_qv(self, key):
        for qv in self.qv_list:
            if qv.name == key:
                return qv()
        raise Exception("Could not find QuantumVariable " + str(key))

    def __str__(self):
        temp_data = list(self.data)
        self.data = []
        i = 0

        from qrisp import QuantumEnvironment

        while temp_data:
            instr = temp_data.pop(i)
            if isinstance(instr, QuantumEnvironment):
                instr.compile()
            else:
                self.append(instr)

        res = "QuantumCircuit:\n---------------\n"

        qc_str = QuantumCircuit.__str__(self)

        # Remove blank line
        lines = qc_str.split("\n")
        if len(set(lines[0])) == 1:
            lines.pop(0)
        # if len(set(lines[-1])) == 1:
        # lines.pop(-1)

        for line in lines:
            res += line + "\n"

        if len(self.env_stack):
            res += "QuantumEnvironment Stack:\n-------------------------\n"
            for i in range(len(self.env_stack)):
                env = self.env_stack[i]
                res += (
                    "Level " + str(i) + ": " + str(type(env)).split(".")[-1][:-2] + "\n"
                )
            res += "\n"

        res += "Live QuantumVariables:\n----------------------"

        qv_name_list = []
        for qv in self.qv_list:
            qv_name_list.append(str(type(qv)).split(".")[-1][:-2] + " " + qv.name)

        for qv_name in qv_name_list:
            res += "\n" + qv_name

        return res

    def get_depth_dic(self):
        return get_depth_dic(self)

    def add_qubit(self, qubit=None):
        qb = super().add_qubit(qubit)
        qb.qs = weakref.ref(self)
        qb.perm_lock = False
        qb.lock = False
        return qb

    def __call__(self):
        return self

    def add_clbit(self, clbit=None):
        cb = super().add_clbit(clbit)
        cb.qs = weakref.ref(self)
        return cb

    def request_qubits(self, request_amount, name=None):
        # Create qubits and add to circuit
        return_qubits = []

        for i in range(request_amount):
            if name is None:
                qb = self.add_qubit()
            else:
                qb = self.add_qubit(Qubit(name + "." + str(i)))

            return_qubits.append(self.qubits[-1])

        for qb in return_qubits:
            self.append(QubitAlloc(), qb)

        return return_qubits

    def clear_qubits(self, qubits, verify=False):
        # Apply initialization operation
        # The following is uncommented because the QASM simulator speed drastically
        # drops when having non unitary operations
        # Uncomment, when executing on real backends
        if not len(self.env_stack):
            pass
            # self.reset(qubits)

        if verify:
            verification_qc = self.copy()
            for qb in qubits:
                verification_qc.measure(qb)

            from qrisp.simulator import run

            res = run(verification_qc, 1000, insert_reset=False)
            for key in res.keys():
                if key[: len(qubits)] != len(qubits) * "0":
                    raise Exception("Tried to delete qubits not in |0> state")

        if not set(qubits).issubset(set(self.qubits)):
            raise Exception(
                "Tried to free up qubits not registered in this quantum session"
            )

        for qb in qubits:
            self.append(QubitDealloc(), qb)

    # Procedure to free up space for quantum variables not used anymore
    def delete_qv(self, qv, verify=False):
        # Check if quantum variable appears in this session
        if qv.name not in [qv.name for qv in self.qv_list]:
            raise Exception(
                "Tried to remove a non existent quantum variable from quantum session"
            )

        self.clear_qubits(qv.reg, verify)

        # Remove quantum variable from list
        for i in range(len(self.qv_list)):
            temp_qv = self.qv_list[i]

            if temp_qv.name == qv.name:
                self.qv_list.pop(i)
                break

        self.deleted_qv_list.append(qv)

    def cnot_count(self):
        """
        Method to determine the amount of CNOT gates used in this QuantumSession.

        Raises
        ------
        Exception
            Tried to compute the CNOT count with open environments.

        Returns
        -------
        int
            The amount of CNOT gates.

        """

        if len(self.env_stack) != 0:
            raise Exception("Tried to count CNOT gates with open if environments")
        from qrisp.misc import cnot_count

        return cnot_count(self)

    def get_local_qvs(self):
        import sys

        local_qvs = []
        for qv in object.__getattribute__(self, "qv_list"):
            if sys.getrefcount(qv) == 3:
                local_qvs.append(qv)

        return local_qvs

        # self.uncomp_stack = []

    def logic_synth(self, input_qubits, output_qubits, tt, method="best", inv=False):
        if len(input_qubits) != tt.bit_amount:
            raise Exception("Given truth table has unfitting amount of input variables")

        if len(output_qubits) != tt.shape[1]:
            raise Exception("Given truth table has unfitting amount of output columns")

        self.append(
            tt.gate_synth(method=method, inv=False), input_qubits + output_qubits
        )

    def depth(self):
        """
        Method to calculate the circuit depth of this QuantumSession

        Returns
        -------
        int
            The circuit depth.

        """
        depth_dic = self.get_depth_dic()
        return max(depth_dic.values())

    def __eq__(self, other):
        return id(self.data) == id(other.data)

    def append(self, operation_or_instruction, qubits=[], clbits=[]):
        # Check the type of the instruction/operation
        from qrisp.circuit import Instruction, Operation

        if issubclass(operation_or_instruction.__class__, Instruction):
            instruction = operation_or_instruction
            self.append(instruction.op, instruction.qubits, instruction.clbits)
            return

        elif issubclass(operation_or_instruction.__class__, Operation):
            operation = operation_or_instruction

        else:
            raise Exception(
                "Tried to append object type "
                + str(type(operation_or_instruction))
                + " which is neither Instruction nor Operation"
            )

        # Convert arguments (possibly integers) to list
        # The logic here is that the list structure gets preserved ie.
        # [[0, 1] ,2] ==> [[qubit_0, qubit_1], qubit_2]
        # unless the input is a single qubit/integer.
        # In this case we have
        # qubit_0 ==> [qubit_0]

        from qrisp.circuit.quantum_circuit import convert_to_cb_list, convert_to_qb_list

        qubits = convert_to_qb_list(qubits, circuit=self)
        clbits = convert_to_cb_list(clbits, circuit=self)

        def check_alloc(input, res=None):
            if isinstance(input, list):
                for item in input:
                    check_alloc(item)
            else:
                if not input.allocated:
                    raise Exception(
                        f"Tried to perform operation {operation.name} on"
                        f"unallocated qubit {input}"
                    )

        if operation.name not in ["qb_alloc", "barrier"]:
            check_alloc(qubits)

        elif operation.name == "qb_alloc":
            qubits[0].allocated = True

        # We now need to merge the sessions and treat their differing environment
        # levels. The idea here is that if a quantum session A is not identical to the
        # environment session B, there have been no gates applied within that
        # environment so far (otherwise merging would have occured). Thus, all data of A
        # belongs into the original_data attribute of the environment with the highest
        # level environment, where the environment quantum session isn't identical to A.

        flattened_qubits = []
        for item in qubits:
            if isinstance(item, Qubit):
                flattened_qubits.append(item)
            else:
                flattened_qubits.extend(item)

        flattened_clbits = []
        for item in clbits:
            if isinstance(item, Clbit):
                flattened_qubits.append(item)
            else:
                flattened_qubits.extend(item)

        # Find the list of all quantum sessions that need to be treated
        qs_list = (
            [qb.qs() for qb in flattened_qubits]
            + [cb.qs() for cb in flattened_clbits]
            + [self]
        )

        # We now iterate through every quantum session and insert its data into the
        # correct original_data attribute
        # for qs in qs_list:

        #     #We need to find the environment where the env_qs quantum session is not merged into qs.
        #     #This implies that the instructions of this session have been appended in this environment's parent.
        #     #Therefore all the data needs to go into the original_data attribute of this environment.
        #     for env in qs.env_stack:
        #         if not env.env_qs == qs:
        #             env.original_data.extend(qs.data)
        #             qs.data = []
        #             merge([qs, env.env_qs])

        # We merge qs_list again since no merge happened incase there were no
        # environments.
        multi_session_merge(qs_list)

        super().append(operation, qubits, clbits)

        if operation.name == "qb_dealloc":
            qubits[0].allocated = False

    def __getitem__(self, key):
        for qv in self.qv_list:
            if qv.name == key:
                return qv
        raise Exception(f"Could not find QuantumVariable {key}")

    # Instead of just resetting the list, we have to use this method.
    # This is because merging two quantum session works essentially by handing them
    # a pointer to the same data list (which contains the merged circuits)
    # If we clear the data list by setting it to an empty list, any session
    # that has been merged with self.qs doesnt point to the same data list anymore.
    # This method tackles this problem by keeping the pointer to the list alive,
    # but removing every single element
    def clear_data(self):
        self.data.clear()

    def statevector(self, return_type="sympy", plot=False, decimals=None):
        r"""
        Returns a representation of the statevector. Three options are available:

        * ``sympy`` returns a `Sympy quantum state
          <https://docs.sympy.org/latest/modules/physics/quantum/state.html>`_,
          which is great for visualization and symbolic investigation. The tensor factors
          are in the order of the creation of the QuantumVariables (or equivalently: as
          they appear, when listed in ``print(self)``).

        * ``latex`` returns the latex code for the Sympy quantum state.

        * ``function`` returns a statevector function, such that the amplitudes can be
          investigated by calling this function on a dictionary of this QuantumSession's
          QuantumVariables.

        If you need to retrieve the statevector as a numpy array, please use the
        corresponding
        :meth:`QuantumCircuit method <qrisp.QuantumCircuit.statevector_array>`.

        Parameters
        ----------
        return_type : str, optional
            String indicating how the statevector should be returned. Available are
            ``sympy``, ``array`` and ``function``. The default is ``sympy``.
        plot : bool, optional
            If the return type is set to ``array``, this boolean will trigger a plot of
            the statevector. The default is ``False``.
        decimals : int, optional
            The decimals to round in the statevector. The default is 5 for return type
            ``sympy`` and infinite otherwise.

        Returns
        -------
        sympy.Expression or LaTeX string or function
            An object representing the statevector.

        Examples
        --------

        We create some QuantumFloats and encode values in them:

        >>> from qrisp import QuantumFloat
        >>> qf_0 = QuantumFloat(3,-1)
        >>> qf_1 = QuantumFloat(3,-1)
        >>> qf_0[:] = 2
        >>> qf_1[:] = {0.5 : 1, 3.5: -1j}

        This encoded the state

        .. math::
            
            \ket{\psi} = \ket{\text{qf_0}} \ket{\text{qf_1}}
            = \frac{1}{\sqrt{2}}  \ket{2} (\ket{0.5} - i \ket{3.5})

        Now we add ``qf_0`` and ``qf_1``:

        >>> qf_res = qf_0 + qf_1

        This gives us the state

        .. math::

            \ket{\phi} = \frac{1}{\sqrt{2}}(\ket{2}\ket{0.5}\ket{2 + 0.5} -
            i \ket{2} \ket{3.5}\ket{2 + 3.5})



        We retrieve the statevector as a Sympy expression:

        >>> sv = qf_0.qs.statevector()
        >>> print(sv)
        sqrt(2)*(|2.0>*|0.5>*|2.5> - I*|2.0>*|3.5>*|5.5>)/4

        If you have Sympy's `pretty printing
        <https://docs.sympy.org/latest/tutorials/intro-tutorial/printing.html>`_ enabled
        in your IPython console, it will even give you a nice Latex rendering:

        >>> sv

        .. image:: ./statevector_print.png
            :width: 300
            :alt: Statevector print
            :align: left

        |
        |

        This feature also works with symbolic parameters:

        >>> from qrisp import QuantumVariable, ry, h, p
        >>> from sympy import Symbol
        >>> qv = QuantumVariable(1)
        >>> ry(Symbol("omega"), qv)
        >>> h(qv)
        >>> p(-Symbol("phi"), qv)
        >>> qv.qs.statevector()

        .. image:: ./symbolic_statevector_print.png
            :width: 350
            :alt: Statevector print
            :align: left

        |
        |

        .. note::

            Statevector simulation with symbolic parameters is significantly more
            demanding than simulation with numeric parameters.

        To retrieve the above expressions as latex code, we use
        ``return_type = "latex"``

        >>> print(qf_0.qs.statevector(return_type = "latex"))
        '\frac{\sqrt{2} \left({\left|2.0\right\rangle }
        {\left|0.5\right\rangle } {\left|2.5\right\rangle }
        - i {\left|2.0\right\rangle } {\left|3.5\right\rangle }
        {\left|5.5\right\rangle }\right)}{2}'


        We can also retrieve the statevector as a Python function:

        >>> sv_function = qf_0.qs.statevector("function")

        Specify the label constellations:

        >>> label_constellation_a = {qf_0 : 2, qf_1 : 0.5, qf_res : 2+0.5}
        >>> label_constellation_b = {qf_0 : 2, qf_1 : 3.5, qf_res : 2+3.5}
        >>> label_constellation_c = {qf_0 : 2, qf_1 : 3.5, qf_res : 4}

        And evaluate the function:

        >>> sv_function(label_constellation_a)
        (0.7071048-1.3411045e-07j)

        This is the expected amplitude up to floating point errors.

        To get a quicker understanding, we can tell the statevector function to round
        the amplitudes using the ``round`` keyword.

        >>> sv_function(label_constellation_b, round = 6)
        (-0-0.707105j)

        Finally, the last amplitude is 0 since the state of ``qf_res`` is not the sum of
        ``qf_0`` and ``qf_1``.

        >>> sv_function(label_constellation_c, round = 6)
        0j


        """

        if len(self.env_stack):
            raise Exception("Tried to evaluate statevector within open QuantumEnvironments")
            
        from qrisp import get_statevector_function, get_sympy_state

        if return_type == "array":
            from qrisp.simulator import statevector_sim

            # Simulate the statevector
            statevector_array = statevector_sim(self.compile())
            # statevector_array = statevector_sim(self)

            # Execute simulation

            # Plot results if required
            if plot:
                import matplotlib.pyplot as plt

                plt.plot(np.real(statevector_array), "o", label="Re(psi)")
                plt.plot(np.imag(statevector_array), "o", label="Im(psi)")
                plt.grid()
                plt.legend()
                plt.show()
            if decimals is None:
                return statevector_array
            else:
                return np.round(statevector_array, decimals)

        elif return_type == "sympy":
            if decimals is None:
                decimals = 6
            return get_sympy_state(self, decimals)

        elif return_type == "latex":
            from sympy import latex

            return latex(self.statevector(return_type="sympy", decimals=decimals))

        elif return_type == "function":
            if decimals is None:
                decimals = 15

            return get_statevector_function(self, decimals)

        else:
            raise Exception(f"Don't know return type {return_type}")

    def compile(
        self,
        workspace=0,
        intended_measurements=[],
        cancel_qfts=True,
        disable_uncomputation=True,
    ):
        r"""
        Method to compile the QuantumSession into a :ref:`QuantumCircuit`. The compiler
        dynamically allocates the qubits of the QuantumSession on qubits that might have
        been used by priorly deleted :ref:`QuantumVariables <QuantumVariable>`.

        Using the ``workspace`` keyword, we can grant the compiler a number of extra
        qubits to use in order to reduce the circuit depth.

        Furthermore, the compiler recompiles any :meth:`mcx <qrisp.mcx>` instruction
        with ``method = auto`` using a dynamically generated mcx implementation that
        makes use of as much of the currently available clean and dirty ancillae.
        This feature will never allocate additional qubits on its own. If required,
        it can be supplied with additional space using the ``workspace`` keyword.

        The .compile method is called by default, when executing the
        :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>` method of
        :ref:`QuantumVariable`. This method also allows specification of compilation
        option through the ``compilation_kwargs`` argument.



        Parameters
        ----------
        workspace : int, optional
            The amount of workspace qubits to be granted. The default is 0.
        intended_measurements : list[Qubit], optional
            A list of :ref:`Qubits <Qubit>` that are supposed to be measured. The
            compiler will remove any instructions that are not directly neccessary to
            perform the measurements. Note that the resulting :ref:`QuantumCircuit`
            contains no measurements, such that the user can still specify a classical
            bit for the measurement. The default ist [].
        cancel_qfts : bool, optional
            If set to True, any :meth:`QFT <qrisp.QFT>` instruction that is executed on
            a set of qubits that have just been allocated (ie. the $\ket{0}$ state) will
            be replaced by a set of H gates. The same goes for QFT instructions executed
            directly before deallocation. The default is ``True``.
        disable_uncomputation : bool, optional
            Experimental feature the allows fully automized uncomputation. If set to
            ``False`` any :ref:`QuantumVariable` that went out of scope will be
            uncomputed by the compiler. The default is ``True``.

        Returns
        -------
        QuantumCircuit
            The compiled QuantumCircuit.

        Examples
        --------

        **Workspace**

        We calculate a product of 2 :ref:`QuantumFloats <QuantumFloat>` using the
        :meth:`sbp_mult <qrisp.sbp_mult>` function which heavily profits from more
        workspace.

        >>> from qrisp import QuantumFloat, sbp_mult
        >>> qf_0 = QuantumFloat(5)
        >>> qf_0[:] = 3
        >>> qf_1 = QuantumFloat(5)
        >>> qf_1[:] = 5

        Calculate product:

        >>> qf_res = sbp_mult(qf_0, qf_1)
        >>> qf_res.qs.num_qubits()
        45

        Compile circuit with no workspace

        >>> qc_0 = qf_res.qs.compile(0)
        >>> qc_0.num_qubits()
        21
        >>> qc_0.depth()
        497

        Compile circuit with 4 workspace qubits

        >>> qc_1 = qf_res.qs.compile(4)
        >>> qc_1.num_qubits()
        25
        >>> qc_1.depth()
        258

        **mcx recompilation**

        To demonstrate the recompilation feature, we create two
        :ref:`QuantumVariables <QuantumVariable>`.

        >>> from qrisp import QuantumVariable, mcx, cx
        >>> ctrl = QuantumVariable(4)
        >>> target = QuantumVariable(1)
        >>> mcx(ctrl, target)
        >>> print(ctrl.qs)
        QuantumCircuit:
        ---------------
          ctrl.0: ──■──
                    │
          ctrl.1: ──■──
                    │
          ctrl.2: ──■──
                    │
          ctrl.3: ──■──
                  ┌─┴─┐
        target.0: ┤ X ├
                  └───┘
        Live QuantumVariables:
        ----------------------
        QuantumVariable ctrl
        QuantumVariable target

        We can now call the ``.compile`` method

        >>> compiled_qc = ctrl.qs.compile()
        >>> compiled_qc.depth()
        50
        >>> print(compiled_qc)
            ctrl.0: ──■──
                      │
            ctrl.1: ──■──
                      │
            ctrl.2: ──■──
                      │
            ctrl.3: ──■──
                    ┌─┴─┐
          target.0: ┤ X ├
                    └───┘

        We see no change here, because there was no free space to execute a more optimal
        mcx implementation. We can grant additional space using the ``workspace``
        argument:

        >>> compiled_qc = ctrl.qs.compile(workspace = 2)
        >>> compiled_qc.depth()
        22
        >>> print(compiled_qc)
                     ┌────────┐               ┌────────┐
             ctrl.0: ┤0       ├───────────────┤0       ├──────────
                     │        │               │        │
             ctrl.1: ┤1       ├───────────────┤1       ├──────────
                     │        │┌────────┐     │        │┌────────┐
             ctrl.2: ┤        ├┤0       ├─────┤        ├┤0       ├
                     │  pt2cx ││        │     │  pt2cx ││        │
             ctrl.3: ┤        ├┤1       ├─────┤        ├┤1       ├
                     │        ││        │┌───┐│        ││        │
           target.0: ┤        ├┤  pt2cx ├┤ X ├┤        ├┤  pt2cx ├
                     │        ││        │└─┬─┘│        ││        │
        workspace_0: ┤2       ├┤        ├──■──┤2       ├┤        ├
                     └────────┘│        │  │  └────────┘│        │
        workspace_1: ──────────┤2       ├──■────────────┤2       ├
                               └────────┘               └────────┘

        Granting extra qubits to use this feature is however not usually necessary. The
        compiler automatically detects and reuses qubit resources available at the
        corresponding stage of the compilation.
        To demonstrate this feature, we allocate a third QuantumVariable:

        >>> qv = QuantumVariable(2)
        >>> cx(target[0], qv)
        >>> print(ctrl.qs.compile())
                  ┌────────┐               ┌────────┐
          ctrl.0: ┤0       ├───────────────┤0       ├────────────────────
                  │        │               │        │
          ctrl.1: ┤1       ├───────────────┤1       ├────────────────────
                  │        │┌────────┐     │        │┌────────┐
          ctrl.2: ┤        ├┤0       ├─────┤        ├┤0       ├──────────
                  │  pt2cx ││        │     │  pt2cx ││        │
          ctrl.3: ┤        ├┤1       ├─────┤        ├┤1       ├──────────
                  │        ││        │┌───┐│        ││        │
        target.0: ┤        ├┤  pt2cx ├┤ X ├┤        ├┤  pt2cx ├──■────■──
                  │        ││        │└─┬─┘│        ││        │┌─┴─┐  │
            qv.0: ┤2       ├┤        ├──■──┤2       ├┤        ├┤ X ├──┼──
                  └────────┘│        │  │  └────────┘│        │└───┘┌─┴─┐
            qv.1: ──────────┤2       ├──■────────────┤2       ├─────┤ X ├
                            └────────┘               └────────┘     └───┘

        We see how the qubits that will later hold ``qv`` are used to efficiently
        compile the mcx gate.

        In situations of no free clean ancilla qubits, the Qrisp compiler even makes use
        of dirty ancillae. To demonstrate, we again create three QuantumVariables
        but this time we execute a :meth:`cx<qrisp.cx>`-gate before executing the
        :meth:`mcx<qrisp.mcx>`-gate. This way ``qv`` has to be allocated before the
        ``mcx`` gate.

        >>> ctrl = QuantumVariable(4)
        >>> target = QuantumVariable(1)
        >>> qv = QuantumVariable(2)
        >>> cx(target[0], qv)
        >>> mcx(ctrl, target)
        >>> print(ctrl.qs.compile())
          ctrl.0: ────────────────────────────────────■──────────────────────────»
                                 ┌─────────────────┐  │  ┌─────────────────┐     »
          ctrl.1: ───────────────┤1                ├──┼──┤1                ├─────»
                                 │                 │  │  │                 │     »
          ctrl.2: ───────────────┤2                ├──┼──┤2                ├─────»
                                 │                 │  │  │                 │     »
          ctrl.3: ────────────■──┤                 ├──┼──┤                 ├──■──»
                            ┌─┴─┐│  reduced_maslov │  │  │  reduced_maslov │┌─┴─┐»
        target.0: ──■────■──┤ X ├┤                 ├──┼──┤                 ├┤ X ├»
                  ┌─┴─┐  │  └─┬─┘│                 │┌─┴─┐│                 │└─┬─┘»
            qv.0: ┤ X ├──┼────┼──┤0                ├┤ X ├┤0                ├──┼──»
                  └───┘┌─┴─┐  │  │                 │└───┘│                 │  │  »
            qv.1: ─────┤ X ├──■──┤3                ├─────┤3                ├──■──»
                       └───┘     └─────────────────┘     └─────────────────┘     »
        «
        «  ctrl.0: ─────────────────────■─────────────────────
        «          ┌─────────────────┐  │  ┌─────────────────┐
        «  ctrl.1: ┤1                ├──┼──┤1                ├
        «          │                 │  │  │                 │
        «  ctrl.2: ┤2                ├──┼──┤2                ├
        «          │                 │  │  │                 │
        «  ctrl.3: ┤                 ├──┼──┤                 ├
        «          │  reduced_maslov │  │  │  reduced_maslov │
        «target.0: ┤                 ├──┼──┤                 ├
        «          │                 │┌─┴─┐│                 │
        «    qv.0: ┤0                ├┤ X ├┤0                ├
        «          │                 │└───┘│                 │
        «    qv.1: ┤3                ├─────┤3                ├
        «          └─────────────────┘     └─────────────────┘

        We see how the qubits of ``qv`` are utilized as dirty ancilla qubits in order
        to facilitate a more efficient ``mcx`` implementation compared to no ancillae
        at all.

        **Fully automized uncomputation**

        This feature is as of right now experimental. To demonstrate, we create a test
        function, creating a local :ref:`QuantumBool` ::

            from qrisp import QuantumBool, mcx

            def triple_AND(a, b, c):

                local = QuantumBool()
                result = QuantumBool()

                mcx([a,b], local)

                mcx([c, local], result)

                return result


        >>> a = QuantumBool()
        >>> b = QuantumBool()
        >>> c = QuantumBool()
        >>> res = triple_AND(a,b,c)
        >>> print(res.qs)
        QuantumCircuit:
        ---------------
             a.0: ──■───────
                    │
             b.0: ──■───────
                    │
             c.0: ──┼────■──
                  ┌─┴─┐  │
         local.0: ┤ X ├──■──
                  └───┘┌─┴─┐
        result.0: ─────┤ X ├
                       └───┘
        Live QuantumVariables:
        ----------------------
        QuantumBool a
        QuantumBool b
        QuantumBool c
        QuantumBool local
        QuantumBool result

        We now compile with the corresponding keyword argument:

        >>> print(a.qs.compile(disable_uncomputation = False))
                     ┌────────┐     ┌────────┐
                a.0: ┤0       ├─────┤0       ├
                     │        │     │        │
                b.0: ┤1       ├─────┤1       ├
                     │        │     │        │
                c.0: ┤  pt2cx ├──■──┤  pt2cx ├
                     │        │┌─┴─┐│        │
           result.0: ┤        ├┤ X ├┤        ├
                     │        │└─┬─┘│        │
        workspace_0: ┤2       ├──■──┤2       ├
                     └────────┘     └────────┘

        We see that the ``local`` QuantumBool is no longer allocated but has been
        uncomputed and it's qubits are available as workspace.


        """
        from qrisp.core.compilation import qompiler

        return qompiler(
            self,
            workspace,
            disable_uncomputation=disable_uncomputation,
            intended_measurements=intended_measurements,
            cancel_qfts=cancel_qfts,
        )

    def __del__(self):
        i = 0
        while i < len(self.qs_tracker):
            if self.qs_tracker[i]() is None or id(self) == id(self.qs_tracker[i]()):
                self.qs_tracker.pop(i)
                continue
            i += 1

    def __hash__(self):
        return id(self.data)

    # The .data attribute is used to identify QuantumSessions with each other even
    # though they are different object. If the .data attribute is set to a new list,
    # this identification is no longer possible because the two different
    # QuantumSessions no longer share the same data list. We overload setattr such that
    # setting a new list results in keeping the old one but with new content.
    def __setattr__(self, name, value):
        if name in ["data"]:
            attr = self.__dict__[name]
            attr.clear()
            attr.extend(value)
        else:
            QuantumCircuit.__setattr__(self, name, value)

    @classmethod
    def get_active_quantum_sessions(self):
        # Remove potential duplicates
        qs_list = list(
            set([qs() for qs in QuantumSession.qs_tracker if not qs() is None])
        )

        self.qs_tracker = [weakref.ref(qs) for qs in qs_list]

        return list(self.qs_tracker)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import networkx as nx
from numba import njit, prange

from qrisp.circuit import QuantumCircuit, Qubit, PTControlledOperation, ControlledOperation, transpile, QubitAlloc, Instruction, fast_append
from qrisp.misc import get_depth_dic, retarget_instructions, reduce_depth


# The purpose of this function is to dynamically (de)allocate qubits when they are
# needed or not needed anymore. The qompiler function knows when a qubit is ready to
# deallocate (ie. it is in |0> state) due to a gate called QubitDealloc. After some
# preprocessing steps, this function will iterate through the data and dynamically
# (de)allocate if it is required. Another feature of this function is that it replaces
# mcx gates where the method is set to "auto" with implementations that fit the
# situation i.e. it takes into consideration how many clean/dirty ancilla qubits are
# available and synthesizes and mcx implementation based on this information.

# The parameter workspace can be given to extend the amount of qubits of the resulting
# QuantumCircuit by that amount. This extension can be helpfull because it gives the
# compiler more options to pick when allocating a new qubit. Since the allocation is
# based on a depth-reduction heuristic, giving more workspace results in many case in
# less depth circuits.


# It is however not only depth that can be improved by an increased workspace. Due to
# more clean/dirty ancillae beeing available, in many cases it is also possible to
# generate more efficient mcx implementations, thus also reducing the gate count.
def qompiler(
    qs,
    workspace=0,
    disable_uncomputation=True,
    intended_measurements=[],
    cancel_qfts=True,
):
    if len(qs.data) == 0:
        return QuantumCircuit(0)

    with fast_append():
        qc = qs.copy()

        if not disable_uncomputation:
            local_qvs = qs.get_local_qvs()

            from qrisp.uncomputation.unqomp import uncompute_qc

            for qv in local_qvs:
                try:
                    qc = uncompute_qc(qc, qv.reg)
                except:
                    print(f"Warning: Automatic uncomputation for {qv.name} failed")
                    

        def reordering_transpile_predicate(op):
            if (
                isinstance(op, PTControlledOperation)
                and op.base_operation.name in ["x"]
            ) or isinstance(op, LogicSynthGate):
                return False
            return True

        from qrisp.logic_synthesis import LogicSynthGate

        if cancel_qfts:
            # The first step is to cancel adjacent QFT gates, which are inverse to each
            # other. This can happen alot because of the heavy use of Fourier arithmetic
            qft_transpile_predicate = (
                lambda op: "QFT" not in op.name and reordering_transpile_predicate(op)
            )

            qc = transpile(qc, transpile_predicate=qft_transpile_predicate)

            qc = qft_cancellation(qc)
            
        if intended_measurements:
            # This function reorders the circuit such that the intended measurements can
            # be executed as early as possible additionally, any instructions that are
            # not needed for the intended measurements are removed
            try:
                qc = measurement_reduction(qc, intended_measurements)
                # pass
            except Exception as e:
                if "Unitary of operation " not in str(e):
                    raise e


        # We now reorder the transpiled QuantumCircuit. Reordering is performed based on
        # the DAG representation of Unqomp. The advantage of this representation is that
        # it abstracts away non-trivial commutation relations between permeable gates.
        # The actual ordering is performed by performing a topological sort on this dag.
        # Contrary to unqomp, we don't use a modified version of Kahns algorithm here
        # (for more information check the topological sort function)

        # The reordering process aims to find an order that minimizes that maximum
        # amount of qubits that is needed. Losely speaking it tries to reorder the data,
        # that allocations are performed as late as possible and deallocations are
        # performed as early as possible.

        # Note that the more aggressive we transpile, the more optimization is possible
        # through reordering i.e. less qubits. By reordering we however also destroy any
        # previous order which might have been intentionally picked to optimize depth.
        # In summary: Transpiling more aggressively leads to less qubits but more depth.

        # Only letting mcx and logic synthesis survive has shown to be a good compromise

        transpiled_qc = transpile(
            qc, transpile_predicate=reordering_transpile_predicate
        )
        reordered_qc = reorder_qc(transpiled_qc)

        # Transpile logic synthesis
        reordered_qc = transpile(
            reordered_qc,
            transpile_predicate=lambda op: not (
                isinstance(op, PTControlledOperation)
                and op.base_operation.name in ["x"]
            ),
        )

        # We combine adjacent single qubit gates
        if not qs.abstract_params:
            reordered_qc = combine_single_qubit_gates(reordered_qc)

        # We now determine the amount of Qubits the circuit will need
        required_qubits = 0
        max_required_qubits = 0

        for instr in reordered_qc.data:
            if instr.op.name == "qb_alloc":
                required_qubits += 1
            elif instr.op.name == "qb_dealloc":
                required_qubits -= 1
            if max_required_qubits < required_qubits:
                max_required_qubits += 1

        # Create a QuantumCircuit. Note that we make sure that the Qubit naming is
        # consistent, since we don't want any non-deterministic elements in the
        # function, as this can hinder bugfixing
        qc = QuantumCircuit()
        for i in range(workspace + max_required_qubits):
            qc.add_qubit(Qubit(str(i)))

        qc.clbits = list(qs.clbits)

        # This dictionary translates between the qubits of the input QuantumSession and
        # the compiled circuit
        translation_dic = {}

        # This list contains the Qubits which are currently not allocated
        free_qb_list = list(qc.qubits)

        allocated_qb_list = []

        depth_dic = {b: 0 for b in qc.qubits + qc.clbits}

        # We now iterate through the data of the preprocessed QuantumCircuit
        for i in range(len(reordered_qc.data)):
            QuantumCircuit.fast_append = True
            instr = reordered_qc.data[i]
            if instr.op.name == "barrier":
                continue

            # Check if any of the involved qubits need an allocation
            for qb in instr.qubits:
                if qb not in translation_dic:
                    # To pick a good allocation, we determine the depth dic and sort the
                    # available Qubits by their corresponding depth. Note that we add
                    # the hash value in order to prevent non-deterministic behavior
                    free_qb_list.sort(
                        key=lambda x: depth_dic[x] + qc.qubits.index(x) * 1e-5
                    )

                    translation_dic[qb] = free_qb_list.pop(0)
                    allocated_qb_list.append(qb)

                    # qc.append(QubitAlloc(), [translation_dic[qb]])

            if instr.op.name == "qb_dealloc":
                # For the deallocation, we simply remove the qubits from the translation
                # dict and append it to the free_qb_list
                free_qb_list.append(translation_dic[instr.qubits[0]])
                allocated_qb_list.remove(instr.qubits[0])

                qc.append(
                    instr.op, [translation_dic[qb] for qb in instr.qubits], instr.clbits
                )

                del translation_dic[instr.qubits[0]]

                continue

            if (
                isinstance(instr.op, ControlledOperation)
                and instr.op.base_operation.name == "x"
                and instr.op.method == "auto"
                and len(instr.qubits) > 3
            ):
                # This section deals with the automatic recompilation of mcx gates.
                # The reason why we do this is, that it might not be clear how many free
                # ancillae are available for mcx gates. If we set the method to auto in
                # the mcx function, this section is called and replaces the generated
                # mcx gate with an implementation that is better fit to suit the amount
                # of available ancillae. We first determine the free ancillae

                clean_ancillae = list(free_qb_list)
                clean_ancillae.sort(
                    key=lambda x: depth_dic[x] + qc.qubits.index(x) * 1e-5
                )

                dirty_ancillae = list(
                    set(translation_dic.values())
                    - set([translation_dic[qb] for qb in instr.qubits])
                )
                dirty_ancillae.sort(
                    key=lambda x: depth_dic[x] + qc.qubits.index(x) * 1e-5
                )

                QuantumCircuit.fast_append = False
                # This function generates the data for the hybrid implementation
                compiled_mcx_data = gen_hybrid_mcx_data(
                    instr.qubits[:-1],
                    instr.qubits[-1],
                    instr.op.ctrl_state,
                    clean_ancillae,
                    dirty_ancillae,
                )

                # We now append the data
                for qb in clean_ancillae + dirty_ancillae:
                    translation_dic[qb] = qb

                for instr in compiled_mcx_data:
                    qc.append(instr.op, [translation_dic[qb] for qb in instr.qubits])
                    update_depth_dic(qc.data[-1], depth_dic)

                # And free up the qubits
                for qb in clean_ancillae + dirty_ancillae:
                    del translation_dic[qb]

                QuantumCircuit.fast_append = True

            # Finally if all of the above cases are not met, we simply append the
            # operation to the translated qubits
            else:
                try:
                    qc.data.append(
                        Instruction(
                            instr.op,
                            [translation_dic[qb] for qb in instr.qubits],
                            instr.clbits,
                        )
                    )

                except KeyError:
                    raise Exception(
                        "Found operation "
                        + instr.op.name
                        + " on unallocated qubit during compilation."
                    )

                update_depth_dic(qc.data[-1], depth_dic)

        # The following code is mostly about renaming and ordering the resulting circuit
        # in order to make the compiled circuit still comprehensible

        # We rename the allocated qubits to their name from the quantum session
        for i in range(len(qs.qubits)):
            if qs.qubits[i] in translation_dic:
                translation_dic[qs.qubits[i]].identifier = qs.qubits[i].identifier
                translation_dic[qs.qubits[i]].hash_value = qs.qubits[i].hash_value

        # We also want the qubits to be sorted according the order of the
        # QuantumVariables, i.e. the order of creation
        sorted_qubit_list = []
        for qv in qs.qv_list:
            # Due to the measurement reduction feature, not all qubits of quantum
            # variables that are live, are guaranteed to be represented in the
            # translation dict. Therefore, we need the try - except structure here
            try:
                sorted_qubit_list.extend([translation_dic[qb] for qb in qv.reg])
            except KeyError:
                pass

        # Furthermore, not all qubits that have been deallocated in the QuantumSession
        # are guaranteed to be removed from the translation dic, since the
        # measurement_reduction function might have removed their
        # uncomputation/deallocation gates
        temp = list(set(translation_dic.values()) - set(sorted_qubit_list))
        temp.sort(key=lambda x: x.identifier)
        sorted_qubit_list.extend(temp)

        # Finally, we rename the deallocated qubits to "workspace"
        workspace_naming_counter = 0
        td_values = list(translation_dic.values())
        for i in range(len(qc.qubits)):
            if not qc.qubits[i] in td_values:
                qc.qubits[i].identifier = "workspace_" + str(workspace_naming_counter)
                workspace_naming_counter += 1
                sorted_qubit_list.append(qc.qubits[i])

        qc.qubits = sorted_qubit_list

        reduced_qc = reduce_depth(qc)

    if reduced_qc.depth() > qc.depth():
        return qc
    else:
        return reduced_qc


def gen_hybrid_mcx_data(controls, target, ctrl_state, clean_ancillae, dirty_ancillae):
    # This function generates the data for the hybrid mcx implementation

    from qrisp.misc.multi_cx import hybrid_mcx
    from qrisp.core import QuantumVariable

    # Specify QuantumVariables to call mcx function
    control_qv = QuantumVariable(len(controls), name="control")
    target_qv = QuantumVariable(1, name="target")

    # dirty_ancillae = dirty_ancillae + clean_ancillae
    hybrid_mcx(
        control_qv,
        target_qv,
        ctrl_state=ctrl_state,
        num_ancilla=len(clean_ancillae),
        num_dirty_ancilla=len(dirty_ancillae),
    )

    # Get the list of used ancillae
    used_ancillae_set = (
        set(control_qv.qs.qubits) - set(control_qv.reg) - set(target_qv.reg)
    )

    # If we used the list() function to transform the set, this introduces a
    # non-deterministic element in the compilation algorithm, which can hamper bugfixing
    used_clean_ancillae = []
    used_dirty_ancillae = []

    for qb in control_qv.qs.qubits:
        if qb in used_ancillae_set:
            if "dirty" in qb.identifier:
                used_dirty_ancillae.append(qb)
            else:
                used_clean_ancillae.append(qb)

    depth_dic = get_depth_dic(control_qv.qs)

    used_clean_ancillae.sort(key=lambda x: -depth_dic[x])
    used_dirty_ancillae.sort(key=lambda x: -depth_dic[x])

    # ancilla_list = list(clean_ancillae)

    # for i in range(len(used_ancillae)):
    #     if "yong" in used_ancillae[i].identifier:
    #         used_ancillae.insert(0, used_ancillae.pop(i))
    #         ancilla_list.insert(0, dirty_ancillae[0])
    #         break

    # Now retarget the instructions such that they use the appropriate qubits
    data = control_qv.qs.data

    retarget_instructions(data, list(control_qv), controls)
    retarget_instructions(data, list(target_qv), [target])
    retarget_instructions(data, used_clean_ancillae, clean_ancillae)
    retarget_instructions(data, used_dirty_ancillae, dirty_ancillae)

    i = 0
    # Remove (de)allocation gates
    while i < len(data):
        if data[i].op.name in ["qb_dealloc", "qb_alloc"]:
            data.pop(i)
            continue
        i += 1

    return data


def reorder_qc(qc):
    from qrisp.uncomputation import dag_from_qc

    G = dag_from_qc(qc, remove_init_nodes=True)

    qc_new = qc.clearcopy()

    dealloc_identifier = lambda x: x.op.name == "qb_dealloc"
    alloc_identifer = lambda x: x.op.name == "qb_alloc"

    # mcx_identifier = lambda x : isinstance(x.op, PTControlledOperation) and x.op.base_operation.name == "x"
    # nmcx_identifier = lambda x : not mcx_identifier(x)
    # sub_sort = lambda G : topological_sort(G, prefer = mcx_identifier, delay = nmcx_identifier)
    # for n in topological_sort(G, prefer = dealloc_identifier, delay = alloc_identifer, sub_sort = sub_sort):

    for n in topological_sort(G, prefer=dealloc_identifier, delay=alloc_identifer):
        qc_new.append(n.instr)

    #The above algorithm does not move allocation gates to their latest possible
    #position (only compared to other deallocation gates)    
    #We therefore manually move the allocation gates to the position right
    #before the first actual instruction on that qubit.
    new_data = []
    delayed_qubit_alloc_dic = {}
    
    for i in range(len(qc_new.data)):
        instr = qc_new.data[i]
        
        if instr.op.name == "qb_alloc":
            delayed_qubit_alloc_dic[instr.qubits[0]] = instr
        else:
            for qb in set(delayed_qubit_alloc_dic.keys()).intersection(instr.qubits):
                new_data.append(delayed_qubit_alloc_dic[qb])
                del delayed_qubit_alloc_dic[qb]
            
            new_data.append(instr)
    
    for instr in delayed_qubit_alloc_dic.values():
        new_data.append(instr)
    
    qc_new.data = new_data

    return qc_new


# This function performs a topological sort of the graph G where we try to execute any
# deallocation gates as early as possible while still adhering to the topological order.
# We try to perform a depth-first search as described here:
# https://en.wikipedia.org/wiki/Topological_sorting
# According to the Wikipedia page, we are allowed to pick any node as a "starting point"
# of the DF-search, which allows us to modify the algorithm such that it optimizes the
# (de)allocation order. The general idea is to pick the deallocation nodes as
# starting points, where we order them, such that those deallocation nodes that
# "require" the least allocation nodes are executed first. "Require" here means that
# there is a causal relationship between the allocation and deallocation nodes,
# i.e. there is a path from the allocation node to the deallocation node.


# We can thus determine the amount of allocation nodes required for a deallocation node
# n by counting, the amount of allocation nodes in the "ancestors" subgraph of n.
def topological_sort(G, prefer=None, delay=None, sub_sort=nx.topological_sort):
    """
    Function to perform a topological sort on an Unqomp DAG which allows preferring/
    delaying specific types of nodes

    Parameters
    ----------
    G : nx.DiGraph
        The Unqomp DAG.
    prefer : function, optional
        Function which returns True, when presented with an Instruction, that should be
        preferred. The default is the function that returns False on all Operations
    delay : function, optional
        Function which returns True, when presented with an Instruction, that should be
        delayed. The default is the function that returns False on all Operations
    sub_sort : function, optional
        A function which performs a topological sort, which can sorting preferences of
        secondary importance. The default is nx.topological_sort.

    Returns
    -------
    lin : list[UnqompNode]
        The linearized list of UnqompNodes. The init nodes are not included.

    """

    if prefer is None:
        prefer = lambda x: False

    if delay is None:
        delay = lambda x: False

    G = G.copy()
    # Collect the prefered nodes
    prefered_nodes = []

    for n in G.nodes():
        if prefer(n.instr):
            prefered_nodes.append(n)

        n.processed = False

    # For large scales, finding the ancestors is a bottleneck. We therefore use a
    # jitted version
    if len(G) * len(prefered_nodes) > 10000:
        anc_lists = ancestors(G, prefered_nodes)
        node_ancs = {
            prefered_nodes[i]: anc_lists[i] for i in range(len(prefered_nodes))
        }
    else:
        node_ancs = {
            prefered_nodes[i]: list(nx.ancestors(G, prefered_nodes[i]))
            for i in range(len(prefered_nodes))
        }
    # Determine the required delay nodes for each prefered nodes
    required_delay_nodes = {n: [] for n in prefered_nodes}

    for n in prefered_nodes:
        for k in node_ancs[n]:
            if delay(k.instr):
                required_delay_nodes[n].append(k)

    required_delay_nodes = {n: set(required_delay_nodes[n]) for n in prefered_nodes}

    # Generate linearization
    lin = []

    while prefered_nodes:
        # Sort nodes accordingly
        prefered_nodes.sort(key=lambda x: len(required_delay_nodes[x]))

        node = prefered_nodes.pop(0)
        ancs = []

        for n in node_ancs[node] + [node]:
            if n.processed:
                continue
            else:
                n.processed = True
                ancs.append(n)

        sub_graph = G.subgraph(ancs)

        lin += list(sub_sort(sub_graph))

        continue

        for n in prefered_nodes:
            required_delay_nodes[n] = (
                required_delay_nodes[n] - required_delay_nodes[node]
            )

    # Linearize the remainder
    remainder = []
    for n in G.nodes():
        if n.processed:
            continue
        else:
            n.processed = True
            remainder.append(n)

    # lin += list(sub_sort(G))
    lin += list(sub_sort(G.subgraph(remainder)))

    return lin


# Function to combine any sequences of single qubit gates into a single U3
def combine_single_qubit_gates(qc):
    def apply_combined_gates(qc_new, gate_list, qb):
        if not len(gate_list):
            return

        n = len(gate_list)

        m = np.eye(2)
        while gate_list:
            gate = gate_list.pop(-1)
            m = np.dot(m, gate.get_unitary())

        if np.linalg.norm(m - np.eye(2)) < 1e-10:
            return

        if n == 1:
            qc_new.append(gate, [qb])
            return

        qc_new.unitary(m, [qb])

    qb_dic = {qb: [] for qb in qc.qubits}

    qc_new = qc.clearcopy()

    for instr in qc.data:
        if (
            len(instr.qubits) > 1
            or instr.op.name in ["qb_alloc", "qb_dealloc"]
            or len(instr.clbits) > 0
        ):
            for qb in instr.qubits:
                apply_combined_gates(qc_new, qb_dic[qb], qb)
            qc_new.append(instr)
        else:
            qb_dic[instr.qubits[0]].append(instr.op)

    for qb in qc.qubits:
        apply_combined_gates(qc_new, qb_dic[qb], qb)

    return qc_new


# This function reorders the circuit such that the intended measurements can be executed
# as early as possible. Additionally, any instructions that are not needed for the
# intended measurements are removed

# Intended measurements has to be a list of qubits

# The strategy is similar to the one presented in reorder_circuit:
# We bring the circuit in the dag representation and perform a topological sort with
# the intended measurement as prefered instructions


# After that we investigate the circuit for instructions that can be removed
def measurement_reduction(qc, intended_measurements):
    qc = qc.copy()

    # Insert intended measurements into the circuit
    for qb in intended_measurements:
        qc.measure(qb)

    # Generate dag representation
    from qrisp.uncomputation import dag_from_qc

    G = dag_from_qc(qc, remove_init_nodes=True)

    # Create result qc
    qc_new = qc.clearcopy()

    # Define prefered instructions
    measure_identifier = (
        lambda x: x.op.name == "measure" and x.qubits[0] in intended_measurements
    )

    # Perform topological sort
    for n in topological_sort(G, prefer=measure_identifier):
        qc_new.append(n.instr)

    # Check which instructions come after the final measurement
    for i in range(len(qc_new.data))[::-1]:
        if measure_identifier(qc_new.data[i]):
            break

    redundant_qc = qc_new.clearcopy()

    redundant_qc.data = qc_new.data[i + 1 :]

    G = dag_from_qc(redundant_qc, remove_init_nodes=True)

    # alloc_graph = allocation_graph(redundant_qc)

    for node in G.nodes():
        if node.instr.op.name == "qb_dealloc":
            ancs = nx.ancestors(G, node)
            for pred in ancs:
                if pred.instr.op.name == "qb_alloc":
                    break
            else:
                redundant_qc.data.remove(node.instr)
                # print(f"removed {node.instr}")
                for pred in ancs:
                    try:
                        redundant_qc.data.remove(pred.instr)
                    except ValueError:
                        pass

    # print(redundant_qc)

    redundant_instructions = redundant_qc.data
    # print(redundant_instructions)

    # redundanct_instructions = qc_new.data[i+1:]
    # #Now we need to make sure we don't remove deallocation gates from the data
    # #because this would inflate the qubit count of the compiled circuit

    # #The strategy here is that if we find a deallocation gate
    # #we remove any instruction involving the deallocated qubit from the list
    # #redundant instructions.

    # #If we however find an allocation gate, this chain of instructions can be
    # #safely removed and therefore stay in the redundant instructions

    # i = 0
    # while i < len(redundant_instructions):
    #     if redundant_instructions[i].op.name == "qb_dealloc":
    #         dealloc_qubit = redundant_instructions[i].qubits[0]

    #         j = 0
    #         while j < len(redundant_instructions):
    #             if dealloc_qubit in redundant_instructions[j].qubits:
    #                 if redundant_instructions[j].op.name == "qb_alloc":
    #                     i += 1
    #                     break

    #                 redundant_instructions.pop(j)
    #                 continue
    #             j += 1
    #         continue
    #     i += 1

    # We now remove the redundant instructions and the inserted
    # measurements from the circuit data
    i = 0
    while i < len(qc_new.data):
        if measure_identifier(qc_new.data[i]):
            qc_new.data.pop(i)
            continue
        if qc_new.data[i] in redundant_instructions:
            qc_new.data.pop(i)
            continue
        i += 1

    return qc_new


def allocation_graph(qc):
    from qrisp.uncomputation.unqomp import dag_from_qc

    dag = dag_from_qc(qc, remove_init_nodes=True)

    res = nx.DiGraph()

    dealloc_nodes = []
    for n in dag.nodes():
        if n.instr.op.name in ["qb_alloc", "qb_dealloc"]:
            res.add_node(n)
        if n.instr.op.name == "qb_dealloc":
            dealloc_nodes.append(n)

    for n in dealloc_nodes:
        for anc in nx.ancestors(n, dag):
            if anc.instr.op.name == "qb_alloc":
                res.add_edge(anc, n)

    return n


def update_depth_dic(instruction, depth_dic):
    if instruction.op.definition:
        qc = QuantumCircuit()
        qc.qubits = instruction.qubits
        qc.append(instruction)
        instr_list = qc.transpile().data
    else:
        instr_list = [instruction]

    # Assign each bit in the circuit a unique integer
    # to index into op_stack.
    # If no bits, return 0

    # Here we are playing a modified version of
    # Tetris where we stack gates, but multi-qubit
    # gates, or measurements have a block for each
    # qubit or cbit that are connected by a virtual
    # line so that they all stacked at the same depth.
    # Conditional gates act on all cbits in the register
    # they are conditioned on.
    # We treat barriers or snapshots different as
    # They are transpiler and simulator directives.
    # The max stack height is the circuit depth.

    for instr in instr_list:
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            continue
        qargs = instr.qubits
        cargs = instr.clbits

        levels = []
        # If count then add one to stack heights

        for b in qargs + cargs:
            # Add to the stacks of the qubits and
            # cbits used in the gate.
            levels.append(depth_dic[b] + 1)

        max_level = max(levels)

        for b in qargs + cargs:
            depth_dic[b] = max_level


@njit(cache=True)
def ancestors_jitted(start_index, indptr, indices, node_amount):
    to_do_array = np.zeros(node_amount, dtype=np.byte)
    to_do_array[start_index] = 1
    done_array = np.zeros(node_amount, dtype=np.byte)

    stack = 1
    while stack:
        node = np.argmax(to_do_array)
        to_do_array[node] = 0

        for i in range(indptr[node], indptr[node + 1]):
            new_node = indices[i]
            if done_array[new_node] == 0:
                to_do_array[new_node] = 1
                stack += 1

        done_array[node] = 1
        stack -= 1

    return np.nonzero(done_array)[0]


@njit(parallel=True, cache=True)
def ancestors_jitted_wrapper(start_indices, indptr, indices, node_amount):
    res = [np.zeros(1, dtype=np.int64)] * len(start_indices)
    for i in prange(len(start_indices)):
        start_index = start_indices[i]
        res[i] = ancestors_jitted(start_index, indptr, indices, node_amount)

    return res


def ancestors(dag, start_nodes):
    node_list = list(dag.nodes())

    sprs_mat = nx.to_scipy_sparse_array(dag, format="csc")

    start_indices = []
    for i in range(len(dag)):
        if node_list[i] in start_nodes:
            start_indices.append(i)

    res_list_indices = ancestors_jitted_wrapper(
        np.array(start_indices).astype(np.int32),
        sprs_mat.indptr,
        sprs_mat.indices.astype(np.int32),
        len(dag),
    )

    node_list = [
        [node_list[j] for j in anc_indices] for anc_indices in res_list_indices
    ]

    return node_list


# Function to cancel adjacent QFT, which are inverse to each other
# Due to the heavy use of Fourier arithmetic, this can happen alot
# especially if multiple arithmetic operation on a single target are executed
def qft_cancellation(qc):
    # The idea is to iterate through the instructions of the circuit
    # and save the last instruction acting on each qubit.
    # If a QFT is appended, we check if the last instruction was an inverse QFT
    # or an allocation and denote the index if so
    # Afterwards, we remove the QFTs or replace them with H gates

    last_instr_dic = {qb: None for qb in qc.qubits}
    cancellation_indices = []
    h_replacements = []
    dealloc_replacements = []
    from numpy.linalg import norm

    for i in range(len(qc.data)):
        
        instr = qc.data[i]
        if "QFT" in instr.op.name:
            previous_instruction_type = []

            for qb in instr.qubits:
                previous_instruction = qc.data[last_instr_dic[qb]]
                
                if previous_instruction.op.num_qubits != instr.op.num_qubits:
                    break
                
                if not qb in previous_instruction.qubits:
                    break
                
                if tuple(previous_instruction.qubits) != tuple(instr.qubits):
                    break

                if qc.data.index(previous_instruction) in cancellation_indices:
                    break

                if "QFT" in previous_instruction.op.name:
                    if instr.op.num_qubits < 8:
                        unitary_self = instr.op.get_unitary()
                        inv_unitary_other = (
                            previous_instruction.op.get_unitary()
                            .transpose()
                            .conjugate()
                        )

                        if bool(norm(unitary_self - inv_unitary_other) < 10**-4):
                            previous_instruction_type.append("QFT")
                        else:
                            break

                    elif hash(instr.op.definition) == hash(
                        previous_instruction.op.definition.inverse()
                    ):
                        previous_instruction_type.append("QFT")
                    else:
                        break

                elif previous_instruction.op.name == "qb_alloc":
                    previous_instruction_type.append("alloc")
                else:
                    break
                

                
            else:
                if len(set(previous_instruction_type)) == 1:
                    if previous_instruction_type[0] == "QFT":
                        cancellation_indices.append(i)
                        cancellation_indices.append(last_instr_dic[qb])
                    else:
                        h_replacements.append(i)

        if instr.op.name == "qb_dealloc":
            deallocated_qubit = instr.qubits[0]
            previous_instruction = qc.data[last_instr_dic[deallocated_qubit]]

            if "QFT" in previous_instruction.op.name:
                try:
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                except AttributeError:
                    previous_instruction.deallocated_qubits = {
                        qb: False for qb in previous_instruction.qubits
                    }
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                    dealloc_replacements.append(last_instr_dic[deallocated_qubit])

        for qb in instr.qubits:
            last_instr_dic[qb] = i

    for i in dealloc_replacements:
        instr = qc.data[i]
        for qb in instr.qubits:
            if not instr.deallocated_qubits[qb]:
                break
        else:
            h_replacements.append(i)

    new_qc = qc.clearcopy()

    # print(len(qc.data))
    for i in range(len(qc.data)):
        if i in h_replacements:
            for qb in qc.data[i].qubits:
                new_qc.h(qb)
            # print("H replacement successfull")
            continue

        if i in cancellation_indices:
            continue
        else:
            new_qc.append(qc.data[i])
    # print(len(new_qc.data))
    # print("====")
    return new_qc



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


def array_entry_manipulation_test():
    from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, multi_measurement

    q_array = QuantumArray(QuantumBool(), (4, 4))
    index_0 = QuantumFloat(2, signed=False)
    index_1 = QuantumFloat(2, signed=False)

    index_0[:] = 2
    index_1[:] = 1

    with q_array[(index_0, index_1)] as entry:
        entry.flip()

    res_array = q_array.most_likely()

    assert res_array[index_0.most_likely(), index_1.most_likely()]



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 22.07.2022
from qrisp import QuantumSession, QuantumArray
from qrisp.arithmetic import QuantumFloat, inplace_matrix_app
import numpy as np


def test_inpl_matrix_multiplication_example():
    def generate_random_inv_matrix(n, bit):
        from qrisp.misc import is_inv
        import random

        found = False

        while found == False:
            matrix = np.zeros((n, n))
            for i in range(n):
                for j in range(n):
                    matrix[i, j] = random.randint(0, 2**bit - 1)

            det = np.round(np.linalg.det(matrix) % 2**bit)

            found = is_inv(det, bit)

        return matrix

    bit = 5
    n = 3
    qs = QuantumSession()

    qtype = QuantumFloat(bit, 0, qs=qs, signed=False)
    vector = QuantumArray(qtype, n)

    x_values = np.array([0, 2, 1])

    vector.encode(x_values)

    matrix = generate_random_inv_matrix(n, bit)

    inplace_matrix_app(vector, matrix)

    print("Matrix multiplication result:")

    expected_res = ((np.dot(matrix, x_values)) % (2**bit)).astype(int).transpose()

    print(expected_res)

    mes_res = vector.get_measurement()
    print("Circuit result:")
    print(mes_res)

    assert list(mes_res.keys())[0] == expected_res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 06.05.2022
from qrisp import QuantumFloat, QuantumArray, tensordot, dot
import numpy as np


def test_matrix_multiplication_example():
    qf = QuantumFloat(3, 0, signed=False)

    test_1 = QuantumArray(qf, shape=(2,))
    test_2 = QuantumArray(qf, shape=(2,))

    test_1[:] = [2, 3]
    test_2[:] = [1, 1]
    # h(test_1)
    # h(test_2)
    print(test_1)
    assert (
        type(test_1[0]).__name__ and type(test_1[1]).__name__
    ) == "QuantumFloat" and type(test_1).__name__ == "QuantumArray"
    assert str(test_1[0]) == "{2: 1.0}" and str(test_1[1]) == "{3: 1.0}"
    # test_1 = QuantumArray(qf, 3)
    # test_1.init_state([(np.array([1,2,3]), 1), (np.array([1,2,2]), 1j)])

    # test_1[:] = np.array([[5,6,2],[3,4,1],[1,1,0],[2,3,5]])

    # test_2 = QuantumArray(qf)
    # test_2[:] = np.eye(3)

    # res = test_1 @ test_2
    # np.sin(test_1)
    res = dot(test_1, test_2)
    print(res)
    assert type(res).__name__ == "QuantumFloat"
    assert str(res) == "{5: 1.0}"
    # test = res.get_measurement(statevector = True, plot = True)

    qf = QuantumFloat(3, 0, signed=False)

    test_1 = QuantumArray(qf, shape=(4, 4))
    test_1.encode(np.eye(4))

    test_1 = test_1.reshape(2, 2, 2, 2)

    test_2 = QuantumArray(qf, shape=(2, 2))
    test_2.encode(np.eye(2))

    res = tensordot(test_1, test_2, (-1, 0))

    qtype = QuantumFloat(3, signed=False)

    test_0 = np.round(np.random.rand(3, 4) * 5)
    test_1 = np.round(np.random.rand(4, 2) * 5)

    q_array = QuantumArray(qtype=qtype)

    q_array[:] = test_1

    test_0 @ q_array

    # test = semi_classic_matrix_multiplication(test_0, q_array)
